import dataclasses
import json
import math
import random
from collections.abc import Callable, Sequence

from lmwrapper.abstract_predictor import LmPredictor
from lmwrapper.structs import LmPrediction, LmPrompt

from synthegrator.code_problems import CodeProblem
from synthegrator.code_solver import ResponseParser
from synthegrator.problem_rendering import (
    LmPromptRender,
    ProblemNotSupportedError,
    PromptRenderer,
    render_markup_to_plain_text,
)
from synthegrator.response_parser import AnswerForSolutionNeverSupported
from synthegrator.transformation_spec import (
    MarkElement,
    SolveStep,
    StsCodeClassificationQuestion,
    StsCodeQuestion,
    StsSpecStatement,
    StsValueQuery,
    markup_path,
)
from synthegrator.util import (
    find_after_first_occurrence_of_string,
    get_only_item,
    normalize_probs_dict,
)

PARSE_CALLBACK = Callable[[LmPrediction], list[SolveStep]]
ANSWER_FOR_SOLUTION_CALLBACK = Callable[[Sequence[SolveStep]], str]


@dataclasses.dataclass(frozen=True)
class PromptRenderWithCallback(LmPromptRender):
    parse_callback: PARSE_CALLBACK
    answer_for_solution_callback: ANSWER_FOR_SOLUTION_CALLBACK = None


class QuestionPromptRenderer(PromptRenderer):
    """
    A prompt renderer for solving StsCodeQuestion problems in
    typical autoregressive way.

    :param randomize_multiple_choice: Whether to randomize the order of the
        multiple choice answers. This might be nice if the model shows
        some bias towards certain answers.
    :param seed_on_problem_id: Only matters if randomize_multiple_choice is
        True. If True, then the randomization will depend based on
        the problem id. Note, that aditional state can still be
        passed in to a make it still random accross runs, but
        a given (problem id, randomizing_seed) pair will always
        produce the same randomization.
    """

    def __init__(
        self,
        randomize_multiple_choice: bool = False,
        seed_on_problem_id: bool = True,
    ):
        # TODO toggle for Chain of Thought
        super().__init__()
        self.randomize_multiple_choice = randomize_multiple_choice
        self.seed_on_problem_id = seed_on_problem_id

    def render(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
        chosen_statement: StsSpecStatement | None = None,
    ) -> LmPromptRender:
        """
        Renders a code problem into a prompt for a language model.
        :param problem: The code problem to render
        :param lm: The language model to render for. Useful if want to
            determine whether to make a dialog or how long can generate.
        :param prompt_seed: A seed to use for any randomization (e.g. for
             few shot seed.)
        :param chosen_statement: The statement to render for in case there are
            multiple questions in the problem. If None, will expect that there
            is exactly one value question in the problem.
        """
        chosen_statement = self._get_and_check_chosen_statement(
            problem,
            chosen_statement,
        )
        try:
            chosen_statement.attempt_to_get_path_as_exact_path()
        except ValueError:
            msg = "Chosen statement must have a path that is a plain path"
            raise ProblemNotSupportedError(
                msg,
            )
        output = ""
        if problem.instructions and problem.instructions_are_essential:
            output += problem.instructions
        output += self._build_code_part(problem, chosen_statement)
        prompt_args, callback, ans_to_sol_cb = self._build_question_part(
            problem,
            chosen_statement,
            prompt_seed,
        )
        output += prompt_args.pop("text")
        return PromptRenderWithCallback(
            prompt=LmPrompt(
                text=output,
                **prompt_args,
            ),
            parse_callback=callback,
            answer_for_solution_callback=ans_to_sol_cb,
        )

    def _build_question_part(
        self,
        problem,
        chosen_statement,
        randomizing_seed,
    ) -> tuple[dict, PARSE_CALLBACK, ANSWER_FOR_SOLUTION_CALLBACK]:
        verb: StsCodeQuestion = chosen_statement.verb
        text = "Question: " + verb.question_text
        prompt_args = {}
        if isinstance(verb, StsCodeClassificationQuestion):
            text += " Answer with only the letter corresponding to the best answer."
            answers = self._get_valid_answers_maybe_randomized(
                verb,
                problem,
                randomizing_seed,
            )
            if len(answers) > 26:
                msg = (
                    "Too many answers for multiple choice question. Expected at most"
                    f" 26. Got {len(answers)}"
                )
                raise ProblemNotSupportedError(
                    msg,
                )
            choice_to_answer = {}
            for i, answer in enumerate(answers):
                letter = chr(ord("A") + i)
                text += f"\n{letter}) {answer}"
                choice_to_answer[letter] = answer
            prompt_args["max_tokens"] = 2
            prompt_args["logprobs"] = 5
            callback, ans_to_sol_cb = self._build_multiple_choice_callback(
                problem,
                chosen_statement,
                choice_to_answer,
            )
        else:
            callback, ans_to_sol_cb = self._build_standard_callback(
                problem, chosen_statement
            )
        text += "\nAnswer:"
        prompt_args["text"] = text
        return prompt_args, callback, ans_to_sol_cb

    def _get_valid_answers_maybe_randomized(
        self,
        verb: StsCodeClassificationQuestion,
        problem: CodeProblem,
        randomizing_seed: int | None,
    ) -> list[str]:
        answers = list(verb.valid_answers)
        if self.randomize_multiple_choice:
            seed = ""
            if self.seed_on_problem_id:
                seed += str(problem.problem_id)
            if randomizing_seed is not None:
                seed += str(randomizing_seed)
            if seed:
                random.seed(seed)
            random.shuffle(answers)
        return answers

    def _build_code_part(self, problem, chosen_statement: StsSpecStatement):
        path = chosen_statement.attempt_to_get_path_as_exact_path()
        markup_root = markup_path(
            problem.working_directory,
            path,
            problem.transformation_spec,
        )
        code_text = render_markup_to_plain_text(
            markup_root,
        )
        lang_spec = problem.get_lang_spec_for_path(
            chosen_statement.attempt_to_get_path_as_exact_path(),
        )
        # fmt: off
        code_text = code_text.strip("\n")
        return (
            f"```{lang_spec.get_lang_md_name()}\n"
            f"{code_text}\n"
            f"```\n"
        )
        # fmt: on

    def _get_and_check_chosen_statement(
        self,
        problem,
        chosen_statement,
    ) -> StsSpecStatement:
        if chosen_statement is None:
            question_nodes = list(problem.transformation_spec.question_statements())
            if len(question_nodes) == 0:
                msg = "No question statements in problem"
                raise ProblemNotSupportedError(msg)
            if len(question_nodes) > 1:
                msg = (
                    "Multiple question statements in problem. Must specify"
                    " chosen_statement."
                )
                raise ProblemNotSupportedError(
                    msg,
                )
            chosen_statement = question_nodes[0]
        if not isinstance(chosen_statement.verb, StsCodeQuestion):
            msg = "chosen_statement must be a StsCodeQuestion"
            raise ValueError(msg)
        return chosen_statement

    def _get_mark_node_for_chosen_statement(
        self,
        problem,
        chosen_statement: StsSpecStatement,
        exception_on_not_found: bool = True,
    ) -> tuple[str, MarkElement[StsValueQuery] | None]:
        path = chosen_statement.attempt_to_get_path_as_exact_path()
        markup = markup_path(
            problem.working_directory,
            path,
            problem.transformation_spec,
        )
        for state in markup.iterate_parser_states():
            if isinstance(state.node.verb, StsValueQuery):
                return path, state.node
        if exception_on_not_found:
            raise RuntimeError
        return path, None

    def _build_standard_callback(
        self,
        problem,
        chosen_statement,
    ) -> PARSE_CALLBACK:
        path, node = self._get_mark_node_for_chosen_statement(problem, chosen_statement)

        def cb(resp: LmPrediction) -> list[SolveStep]:
            resp_str, token = grab_after_prompt(resp, "Answer:")
            return [
                SolveStep(
                    path,
                    node.mark_id,
                    value=resp_str,
                ),
            ]

        def answer_to_solution_cb(solve_steps: list[SolveStep]) -> str:
            val = get_only_item(solve_steps).value
            if not val.startswith(" "):
                return " " + val
            return val

        return cb, answer_to_solution_cb

    def _build_multiple_choice_callback(
        self,
        problem,
        chosen_statement,
        choice_to_answer: dict[str, str],
        case_insensitive: bool = True,
        strip_whitespace: bool = True,
    ) -> tuple[PARSE_CALLBACK, ANSWER_FOR_SOLUTION_CALLBACK]:
        path, node = self._get_mark_node_for_chosen_statement(problem, chosen_statement)
        new_choice_to_answer = dict(choice_to_answer)
        answer_to_choice = {v: k for k, v in new_choice_to_answer.items()}
        #print("Answer to choice")
        #print(answer_to_choice)
        if case_insensitive:
            new_choice_to_answer = {
                k.lower(): v for k, v in new_choice_to_answer.items()
            }
        if strip_whitespace:
            new_choice_to_answer = {
                k.strip(): v for k, v in new_choice_to_answer.items()
            }
        available_answers = list(new_choice_to_answer.values())
        # Make an actual answers map to itself
        new_choice_to_answer |= {v: v for v in new_choice_to_answer.values()}
        #print("New choice to answer")
        #print(new_choice_to_answer)

        def cb(resp: LmPrediction) -> list[SolveStep]:
            resp_str, token_index_start = grab_after_prompt(resp, "Answer:")
            token_log_probs = resp.top_token_logprobs[token_index_start]
            #print("Token log probs")
            #print(token_log_probs)
            #exit()
            answer_to_prob = {}
            for tok, logprob in token_log_probs.items():
                if strip_whitespace:
                    tok = tok.strip()
                if case_insensitive:
                    tok = tok.lower()
                if tok in new_choice_to_answer:
                    answer = new_choice_to_answer.get(tok, None)
                    if answer is None:
                        continue
                    if answer in answer_to_prob:
                        answer_to_prob[answer] += math.exp(logprob)
                    else:
                        answer_to_prob[answer] = math.exp(logprob)
            answer_to_prob = normalize_probs_dict(
                answer_to_prob,
                available_answers,
            )
            return [
                SolveStep(
                    path,
                    node.mark_id,
                    value=json.dumps(answer_to_prob),
                ),
            ]

        verb: StsCodeClassificationQuestion = chosen_statement.verb

        def answer_to_solution_cb(solve_steps: list[SolveStep]) -> str:
            answer_to_prob = verb.value_to_prob_dict(get_only_item(solve_steps).value)
            answer = max(answer_to_prob, key=answer_to_prob.get)
            if strip_whitespace:
                return " " + answer_to_choice[answer]
            return answer_to_choice[answer]

        return cb, answer_to_solution_cb


def grab_after_prompt(
    resp: LmPrediction,
    prompt_str: str,
) -> tuple[str, int]:
    if prompt_str not in resp.completion_text:
        return resp.completion_text, 0
    token_offset, within_token_offset = find_after_first_occurrence_of_string(
        resp.completion_tokens,
        prompt_str,
    )
    if within_token_offset != 0:
        msg = "Not currently dealing with weird split tokens"
        raise NotImplementedError(msg)
    return resp.completion_text.split(prompt_str, 1)[1], token_offset


def res_to_str(
    resp: LmPrediction,
) -> str:
    return resp.completion_text


class CallbackResponseParser(ResponseParser):
    def __init__(self):
        super().__init__()

    def parse(
        self,
        render: PromptRenderWithCallback,
        resp: LmPrediction,
        problem: CodeProblem,
    ) -> list[SolveStep]:
        return render.parse_callback(resp)

    def answer_for_solution(
        self,
        render: PromptRenderWithCallback,
        problem: CodeProblem,
        solve_steps: Sequence[SolveStep],
    ) -> str:
        if render.answer_for_solution_callback is None:
            raise AnswerForSolutionNeverSupported()
        return render.answer_for_solution_callback(solve_steps)



import structlog
import structlog._config

__all__ = ["Profile", "configure"]


IS_CONFIGURED = False
PROFILES = {}
DOCKER_URL = None
DOCKER_TIMEOUT = 30
DOCKER_MAX_TOTAL_RETRIES = 9
DOCKER_MAX_CONNECT_RETRIES = 5
DOCKER_MAX_READ_RETRIES = 5
DOCKER_BACKOFF_FACTOR = 0.2
DOCKER_WORKDIR = "/sandbox"

DEFAULT_LIMITS = {
    # CPU time in seconds, None for unlimited
    "cputime": 1,
    # Real time in seconds, None for unlimited
    "realtime": 5,
    # Memory in megabytes, None for unlimited
    "memory": 64,
    # limit the max processes the sandbox can have
    # -1 or None for unlimited(default)
    "processes": -1,
}
DEFAULT_USER = "root"
CPU_TO_REAL_TIME_FACTOR = 5


class Profile:
    def __init__(
        self,
        name,
        docker_image,
        command=None,
        user=DEFAULT_USER,
        read_only=False,
        network_disabled=True,
    ):
        self.name = name
        self.docker_image = docker_image
        self.command = command
        self.user = user
        self.read_only = read_only
        self.network_disabled = network_disabled


def configure(profiles=None, docker_url=None):
    global IS_CONFIGURED, PROFILES, DOCKER_URL

    IS_CONFIGURED = True
    if isinstance(profiles, dict):
        profiles_map = {
            name: Profile(name, **profile_kwargs)
            for name, profile_kwargs in profiles.items()
        }
    else:
        profiles_map = {profile.name: profile for profile in profiles or []}
    PROFILES.update(profiles_map)
    DOCKER_URL = docker_url


# structlog.is_configured() was added in 18.1
if not structlog._config._CONFIG.is_configured:
    structlog.configure(
        processors=[
            structlog.stdlib.filter_by_level,
            structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.KeyValueRenderer(key_order=["event"]),
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )



from typing import Dict, Any, Optional, Union, Iterator, ClassVar
from synthegrator.code_problems import CodeProblem
from synthegrator.util import IteratorWithLength
import logging
from typing import Optional, Union, Dict, List, Any

logger = logging.getLogger(__name__)

class DatasetNameType:
    """Base class for dataset name type to use in type annotations"""
    pass



class DatasetSpec(DatasetNameType):
    """Specification for a dataset with optional filters."""
    
    def __init__(
        self,
        name: str,
        display_name: Optional[str] = None,
        is_a_base_collection: bool = False,
        parent: Optional['DatasetSpec'] = None,
        **filters: Any
    ) -> None:
        """
        Initialize a dataset specification.
        
        Args:
            name: The name of the dataset (e.g., "humaneval_plus").
            display_name: Optional friendly display name for the dataset (e.g., "HumanEval+")
            is_a_base_collection: Whether this dataset is a distinct collection from a 
                                 specific source/paper, not just a filtered view
            parent: Optional parent specification to inherit filters from
            **filters: Filter key-value pairs to apply to this dataset

        The actual items that get yielded by dataset should be the
        get_base_name(). So something like mbpp_sanatized will yield
        items with a dataset_name of "mbpp" which is the name of it's
        base collection parent.
        """
        self.name = name
        self.display_name = display_name or name
        self.is_a_base_collection = is_a_base_collection
        self.parent = parent
        self.filters = filters
        
    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DatasetSpec):
            return (self.name == other.name and
                    self.is_a_base_collection == other.is_a_base_collection and
                    self.filters == other.filters and
                    (self.parent == other.parent if self.parent and other.parent else
                     self.parent is other.parent))
        elif isinstance(other, str):
            return self.name == other
        return NotImplemented

    def __hash__(self) -> int:
        # Create a hashable representation of filters
        filters_hash = frozenset(self.filters.items())
        # Include parent name instead of parent object to avoid recursion
        parent_hash = self.parent.name if self.parent else None
        return hash((self.name, parent_hash, filters_hash, self.is_a_base_collection))

    def __repr__(self) -> str:
        """Create a deterministic string representation for caching"""
        parent_str = f", parent={self.parent.name}" if self.parent else ""
        filters_str = ", " + ", ".join(f"{k}={v!r}" for k, v in self.filters.items()) if self.filters else ""
        return f"DatasetSpec(name={self.name!r}, is_base={self.is_a_base_collection}{parent_str}{filters_str})"

    @property
    def is_root(self) -> bool:
        """Whether this is a root dataset specification with no parent."""
        return self.parent is None
        
    def get_base_name(self) -> str:
        """Get the base name for routing to the appropriate dataset loader."""
        if not self.is_a_base_collection and self.parent:
            return self.parent.get_base_name()
        return self.name

    def get_all_filters(self) -> Dict[str, Any]:
        """Get all filters, including parent filters."""
        if not self.parent:
            return self.filters.copy()
            
        all_filters = self.parent.get_all_filters()
        all_filters.update(self.filters)
        return all_filters
        
    def get_base_collection(self) -> 'DatasetSpec':
        """Get the nearest ancestor (including self) that is a base collection."""
        if self.is_a_base_collection:
            return self
        elif self.parent:
            return self.parent.get_base_collection()
        # Fallback in case there's no base collection in the hierarchy
        return self
        
    def get_base_collection_name(self) -> str:
        """Get the name of the nearest base collection in the hierarchy."""
        base_collection = self.get_base_collection()
        return base_collection.name


class DatasetNameMeta(type):
    def __iter__(cls):
        """Make the DatasetName class directly iterable."""
        for attr_name, attr_value in vars(cls).items():
            if isinstance(attr_value, DatasetSpec) and not attr_name.startswith('_'):
                yield attr_value

    def __instancecheck__(cls, instance):
        """Make isinstance(x, DatasetName) work for DatasetSpec instances defined in DatasetName."""
        if isinstance(instance, DatasetSpec):
            # Check if this instance is one of the dataset specs defined in DatasetName
            for attr_name, attr_value in vars(cls).items():
                if not attr_name.startswith('_') and instance == attr_value:
                    return True
        return False


class DatasetName(DatasetNameType, metaclass=DatasetNameMeta):
    """Collection of available datasets and their variants."""
    # TODO: add some way of registering datasets and their loaders
    # dymanically so easier for downstream libraries.

    # Base datasets - these are all distinct collections from papers/sources
    humaneval: ClassVar[DatasetSpec] = DatasetSpec(
        name="humaneval", 
        display_name="HumanEval",
        is_a_base_collection=True
    )
    
    mbpp: ClassVar[DatasetSpec] = DatasetSpec(
        name="mbpp",
        display_name="MBPP",
        is_a_base_collection=True,
        format_as_method_completion_problem=True,
    )
    
    livecodebench: ClassVar[DatasetSpec] = DatasetSpec(
        name="livecodebench", 
        display_name="LiveCodeBench",
        is_a_base_collection=True,
    )
    
    dypy_line_completion: ClassVar[DatasetSpec] = DatasetSpec(
        name="dypy_line_completion",
        display_name="DyPy",
        is_a_base_collection=True
    )
    
    humaneval_plus: ClassVar[DatasetSpec] = DatasetSpec(
        name="humaneval_plus",
        display_name="HumanEval+", 
        is_a_base_collection=True,  # It's a distinct collection from eval+
        parent=humaneval,  # But inherits routing information
        use_eval_plus=True
    )
    
    mbpp_plus: ClassVar[DatasetSpec] = DatasetSpec(
        name="mbpp_plus",
        display_name="MBPP+", 
        is_a_base_collection=True,  # It's a distinct collection from distinct paper
        parent=mbpp,  # But inherits routing information
        format_as_method_completion_problem=True,
    )
    
    # Just a filtered view of MBPP
    mbpp_sanatized: ClassVar[DatasetSpec] = DatasetSpec(
        name="mbpp_sanatized",
        display_name="MBPP-Sanitized", 
        is_a_base_collection=False,  # Not a distinct collection, just a filtered view
        parent=mbpp,
        sanitized_only=True,
    )
    
    # Filtered views of LiveCodeBench by difficulty
    livecodebench_easy: ClassVar[DatasetSpec] = DatasetSpec(
        name="livecodebench_easy",
        display_name="LiveCodeBench-Easy", 
        is_a_base_collection=False,
        parent=livecodebench,
        filter_difficulty="easy"
    )
    
    livecodebench_medium: ClassVar[DatasetSpec] = DatasetSpec(
        name="livecodebench_medium",
        display_name="LiveCodeBench-Medium", 
        is_a_base_collection=False,
        parent=livecodebench,
        filter_difficulty="medium"
    )
    
    livecodebench_hard: ClassVar[DatasetSpec] = DatasetSpec(
        name="livecodebench_hard",
        display_name="LiveCodeBench-Hard", 
        is_a_base_collection=False,
        parent=livecodebench,
        filter_difficulty="hard"
    )
    

    repocod: ClassVar[DatasetSpec] = DatasetSpec(
        name="repocod",
        display_name="REPOCOD", 
        is_a_base_collection=True,
    )


# Type for the dataset parameter 
DatasetSpecifier = Union[str, DatasetSpec]


def yield_problems_from_name(
    dataset: DatasetSpecifier,
    max_problems: Optional[int] = None,
    **kwargs: Any
) -> Union[Iterator[CodeProblem], IteratorWithLength[CodeProblem]]:
    """
    Load and yield problems from a dataset by name, with optional filtering.
    
    Args:
        dataset: A dataset specifier - either a string name or DatasetSpec object
        max_problems: Maximum number of problems to yield
        **kwargs: Additional arguments to pass to the dataset yield function
        
    Returns:
        An iterator of CodeProblem instances
        
    Examples:
        > yield_problems_from_name(DatasetName.LIVECODEBENCH_EASY, max_problems=10)
        > yield_problems_from_name("humaneval", use_eval_plus=True)
        > yield_problems_from_name("mbpp", sanitized_only=True)
    """
    if isinstance(dataset, DatasetSpec):
        # Extract base dataset name and filters
        base_name = dataset.get_base_name()
        filters = dataset.get_all_filters()
        
        # Merge filters with kwargs (kwargs take precedence)
        merged_kwargs = filters.copy()
        merged_kwargs.update(kwargs)
        kwargs = merged_kwargs
    else:
        base_name = str(dataset)
        
    # Pass max_problems to all dataset loaders
    if max_problems is not None:
        kwargs['max_problems'] = max_problems
    
    # Dispatch to the appropriate dataset loader
    if base_name == DatasetName.humaneval or base_name == DatasetName.humaneval_plus:
        from synthegrator.synthdatasets.human_eval import yield_human_eval
        return yield_human_eval(**kwargs)
    elif base_name == DatasetName.mbpp:
        from synthegrator.synthdatasets.mbpp import yield_mbpp
        return yield_mbpp(**kwargs)
    elif base_name == DatasetName.mbpp_plus:
        from synthegrator.synthdatasets.mbpp import yield_mbpp_plus
        return yield_mbpp_plus(**kwargs)
    elif base_name == DatasetName.dypy_line_completion:
        from synthegrator.synthdatasets.dypybench import yield_dypybench
        return yield_dypybench(**kwargs)
    elif base_name == DatasetName.livecodebench:
        from synthegrator.synthdatasets.livecodebench import yield_livecode_problems
        return yield_livecode_problems(**kwargs)
    elif base_name == DatasetName.repocod:
        from synthegrator.synthdatasets.repocod import yield_repocod_problems
        return yield_repocod_problems(**kwargs)
    else:
        raise ValueError(f"Unknown dataset: {base_name}")


def load_hf_dataset_with_backoff(
    path: str,
    name: Optional[str] = None,  # This should always be specified
    max_retries: int = 5,
    initial_delay: float = 1.0,
    max_delay: float = 60.0,
    jitter: float = 0.1,
    **kwargs
):
    """
    Load a Hugging Face dataset with exponential backoff for handling rate limits.

    Args:
        path: Dataset path or name
        name: Dataset configuration name (SHOULD BE EXPLICITLY SPECIFIED)
        max_retries: Maximum number of retry attempts
        initial_delay: Initial delay between retries in seconds
        max_delay: Maximum delay between retries in seconds
        jitter: Random jitter factor to add to delay
        **kwargs: Additional arguments to pass to load_dataset
    """
    import time
    import random
    from datasets import load_dataset
    from huggingface_hub.utils import RepositoryNotFoundError, RevisionNotFoundError
    from requests.exceptions import HTTPError, ConnectionError, Timeout
    retry_exceptions = (
        HTTPError,
        ConnectionError,
        Timeout,
        FileNotFoundError,
    )

    last_exception = None
    delay = initial_delay

    for attempt in range(max_retries):
        try:
            # Always use the explicitly provided configuration
            if name is not None:
                return load_dataset(path, name, **kwargs)
            else:
                return load_dataset(path, **kwargs)

        except retry_exceptions as e:
            last_exception = e

            # Only handle rate limiting and connection errors
            if attempt < max_retries - 1:
                # Calculate sleep time with exponential backoff and jitter
                jitter_amount = random.uniform(-jitter, jitter) * delay
                sleep_time = min(delay + jitter_amount, max_delay)

                time.sleep(sleep_time)
                delay = min(delay * 2, max_delay)

    if last_exception:
        raise last_exception
    else:
        raise RuntimeError(f"Failed to load dataset {path} after {max_retries} attempts")


from synthegrator.synthdatasets.livecodebench import yield_livecode_problems
from synthegrator.code_problems import TestCase, TestCase, TestCaseArbitraryMethod
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec

def main():
    probs = yield_livecode_problems(max_problems=100)
    invalid_tcs = []
    for prob in probs:
        print(dir(prob))
        for tc in prob.test_cases:
            print(type(tc))
            formatted_tc = tc.format_for_test_framework(lang_spec=PythonLangSpec())
            # Assert that formatted_tc is compilable python code
            try:
                compile(formatted_tc, '<string>', 'exec')
            except SyntaxError as e:
                #raise AssertionError(f"Formatted test case is not valid Python code: {e}") from e
                invalid_tcs.append(prob)
                break
            if "LCB STDIN TEST CASE" in formatted_tc:
                continue
            if len(formatted_tc) > 500:
                formatted_tc = formatted_tc[:500] + "..."
            print(formatted_tc)
            break
    print(len(invalid_tcs))
            

if __name__ == "__main__":
    main()


import ast
import doctest
import logging
logger = logging.getLogger(__name__)

from synthegrator.code_problem_builders import make_simple_method_completion_problem
from synthegrator.code_problems import (
    CodeProblem,
    TestCase,
    TestCaseArbitraryMethod,
    TestCaseExpressionIsEq,
    TestCaseMethodCallIsEq,
)
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.solution_eval import apply_solution
from synthegrator.util import IteratorWithLength


def produce_code_problem_for_human_eval(
    task_id: str,
    prompt: str,
    canonical_solution: str,
    test: str,
    entry_point: str,
    is_eval_plus: bool,
) -> CodeProblem:
    imports, method = PythonLangSpec.split_before_last_method_name(prompt)
    signature, prompt = method.split("\n", 1)
    # ^ human eval uses irregular test case style. Some are real doctests, some are
    #   just text in an 'example' section. Some are doctests but done wrong (like are == expression)
    test_cases = _parse_assert_style_test_cases(test, entry_point)

    _original_prompt = prompt
    prob = make_simple_method_completion_problem(
        prompt=prompt,
        signature=signature,
        test_cases=test_cases,
        model_context=imports,
        known_solutions=[canonical_solution],
        lang_spec=PythonLangSpec(),
        prompt_already_docstring=True,
        problem_id=task_id,
        dataset_name="humaneval" if not is_eval_plus else "humaneval_plus",
    )
    apply_solution(prob.known_solutions[0])
    return prob


def yield_human_eval(
    use_eval_plus: bool = False,
    max_problems: int = None,
) -> IteratorWithLength[CodeProblem]:
    """
    Yields code problems from the HumanEval dataset.

    Args:
    ----
        use_eval_plus: If True, use the eval+ (from https://evalplus.github.io/)
            dataset instead of the original HumanEval dataset. This expands the
            number of test cases.

    """
    from datasets import load_dataset

    if not use_eval_plus:
        dataset = load_dataset("openai_humaneval")
    else:
        dataset = load_dataset("evalplus/humanevalplus")
    data = dataset["test"]

    def gen():
        for ex in data:
            yield produce_code_problem_for_human_eval(
                task_id=ex["task_id"],
                prompt=ex["prompt"],
                canonical_solution=ex["canonical_solution"],
                test=ex["test"],
                entry_point=ex["entry_point"],
                is_eval_plus=use_eval_plus,
            )

    return IteratorWithLength(gen(), min(len(data), max_problems or 9e9))


def _get_docstring_test_in_outs(docstring: str) -> list[tuple[str, str]]:
    docstring = docstring.strip()[3:-3]
    parser = doctest.DocTestParser()

    # Parse the doctests from the docstring
    doc_test = parser.get_doctest(docstring, {}, "test", "test", 0)

    # Extract and return the individual test cases
    test_cases = []
    for example in doc_test.examples:
        source = example.source.strip()
        want = example.want.strip()
        test_cases.append((source, want))
    return test_cases


def _parse_doctest_style_test_cases(docstring: str, method_name: str) -> list[TestCase]:
    doc_test_in_outs = _get_docstring_test_in_outs(docstring)
    test_cases = []
    for i, (source, want) in enumerate(doc_test_in_outs):
        is_test_method_call = source.startswith(f"{method_name}(") and source.endswith(
            ")",
        )
        output = ast.literal_eval(want) if want.strip() else None
        if is_test_method_call:
            args = ast.literal_eval(source[len(method_name) + 1 : -1])
            test_cases.append(
                TestCaseMethodCallIsEq(
                    is_hidden_test=False,
                    fail_message=f"Fail docstring test method call case {i}: {source}",
                    method_name=method_name,
                    test_input=args,
                    output=output,
                ),
            )
        else:
            test_cases.append(
                TestCaseExpressionIsEq(
                    is_hidden_test=False,
                    fail_message=f"Fail docstring test expression case {i}: {source}",
                    input_expression=source,
                    output=output,
                ),
            )
    return test_cases


def _parse_assert_style_test_cases(code: str, test_method_name: str) -> list[TestCase]:
    # TODO split several asserts into different test cases
    # pull out just the body
    code += f"\ncheck({test_method_name})"
    return [
        TestCaseArbitraryMethod(
            is_hidden_test=True,
            fail_message=None,
            method_body=code,
            lang_spec=PythonLangSpec(),
        ),
    ]


def main():
    exs = next(yield_human_eval())
    logger.info(exs)


if __name__ == "__main__":
    main()



try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version")

# Configure package-wide logging
import logging

# Disable DEBUG logs from urllib3
logging.getLogger('urllib3').setLevel(logging.INFO)

# Create the root package logger
logger = logging.getLogger("synthegrator")

# Only configure if not already set up
if not logger.handlers:
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s [%(name)s:%(lineno)d] %(levelname)s: %(message)s')
    formatter.datefmt = '%H:%M:%S'  # Just hours:minutes:seconds
    handler.setFormatter(formatter)
    logger.setLevel(logging.INFO)
    logger.addHandler(handler)



import platform
import warnings
from collections.abc import Iterator
from pathlib import Path

import pandas as pd

from synthegrator.code_problem_builders import make_simple_line_edit_problem
from synthegrator.code_problems import (
    CodeProblem,
    DiscoveredTestsuite,
)
from synthegrator.environments import CodeExecutionEnvironment
from synthegrator.lang_specs.lang_spec_java import JavaLangSpec
from synthegrator.sandboxing import Cmd, DockerExecutionContext, ExecLimits
from synthegrator.transformation_spec import (
    StsEditable,
    StsSelectorLineSpan,
    StsSpecStatement,
    TransformationSpec,
)
from synthegrator.util import (
    MultilineString,
)

_defects4j_root = Path(__file__).parent / "defects4j_data"


def _default_projects_root() -> Path:
    directory = _defects4j_root / "Defects4J_projects_clean"
    if not directory.exists():
        if (_defects4j_root / "Defects4J_projects_clean.zip").exists():
            # Unzip the projects
            # TODO eventually just read directly from the zip file
            import zipfile

            with zipfile.ZipFile(
                _defects4j_root / "Defects4J_projects_clean.zip", "r"
            ) as zip_ref:
                zip_ref.extractall(_defects4j_root)
    assert directory.exists()
    return directory


def _make_docker_execution_context():
    dockerfile = (_defects4j_root / "Dockerfile").read_text()
    is_arm = "arm" in platform.machine().lower()
    if is_arm:
        dockerfile.replace(
            "openjdk-amd64",
            "openjdk-arm64",
        )
    return DockerExecutionContext(
        image_name="synthegrator_defects4j",
        dockerfile=dockerfile,
        build_pre_cmds=Cmd.from_cmds_str_mix_tuple(
            f"cp -r {_defects4j_root}/Defects4J_projects_clean .",
            f"cp {_defects4j_root}/dataset.py .",
            f"cp {_defects4j_root}/defects4j.build.xml .",
            f"cp {_defects4j_root}/Defects4J_oneLiner_metadata.csv .",
        ),
        default_limits=ExecLimits(
            timeout_cpu_s=60 * 10,
            timeout_realtime_s=60 * 20,
            networking_allowed=False,
        ),
    )


DEFECTS_4_J_DOCKER_ENV = _make_docker_execution_context()


def _figure_out_known_solution_for_defects4j(
    project_name: str,
    problem_id: str,
    file_name: str,
    line_no: int,
):
    # Diff the two paths
    # Render the problem markup so that way we know get a mark_id for the node
    # assert just one diff line
    # add a solve step for the whatever the mark_id is and new line value
    # (I can help explain this a bit better)
    solution_path = (
        _default_projects_root() / f"{project_name}_{problem_id}_fixed" / file_name
    )
    return MultilineString(solution_path.read_text())[line_no]


INSTRUCTIONS_ONE_LINE = (
    "There is a single line bug in this java code. " "Your task is to repair the bug."
)


def make_problem_for_defect4j(path, line_no):
    problem = CodeProblem(
        instructions=(
            "There is a single line bug in this java code. "
            "Your task is to repair the bug."
        ),
        # instructable_metadata={
        # InstructableMetadataKeys.commit_message: "",  # commit_message
        # },
        transformation_spec=_make_transformation_spec_single_line(path, line_no),
        environment=CodeExecutionEnvironment(docker_env=DEFECTS_4_J_DOCKER_ENV),
    )
    problem.known_solutions = _figure_out_known_solution_for_defects4j()
    return problem


def _make_transformation_spec_single_line(path, line_no):
    return TransformationSpec(
        statements=(
            StsSpecStatement(
                path,
                StsSelectorLineSpan(line_no - 1, line_no),
                StsEditable(current_text_visible=True, lang_spec_name="java"),
            ),
        ),
    )


def make_discovered_test_case(
    project_name: str,
    problem_id: str,
    file_path: str,
    is_buggy: bool = True,  # Run the buggy version by default since that's the one we'll patch
):
    # s1 = "ls /usr/lib/jvm/java-8-openjdk-arm64"
    # s2 = "ls /usr/lib/jvm/java-8-openjdk-amd64"
    # assert s1 == s2
    return DiscoveredTestsuite(
        cmds=[
            Cmd(
                f"cd /defects4j/Defects4J_projects/{project_name}_{problem_id}{'' if is_buggy else '_fixed'} &&"
                f" rm /defects4j/Defects4J_projects/{project_name}_{problem_id}{'' if is_buggy else '_fixed'}/{file_path} &&"
                " cp /sandbox/solution.java"
                f" /defects4j/Defects4J_projects/{project_name}_{problem_id}{'' if is_buggy else '_fixed'}/{file_path} &&"
                " defects4j test && cp TESTS-TestSuites.xml"
                " /sandbox/TESTS-TestSuites.xml",
                files=(
                    {
                        "name": "defects4j.build.xml",
                        "content": (
                            _defects4j_root / "defects4j.build.xml"
                        ).read_bytes(),
                    },
                ),
            ),
            # Cmd(
            #     f"cd /defects4j/Defects4J_projects/{project_name}_{problem_id}{'' if is_buggy else '_fixed'} && defects4j test"
            # ),
            # Cmd(f"ls /defects4j/Defects4J_projects/Chart_1_fixed"),
            # Cmd(
            #    f"cp /sandbox/defects4j.build.xml /defects4j/Defects4J_projects/Chart_1_fixed/defects4j.build.xml"
            #    f" && cd /defects4j/Defects4J_projects/Chart_1_fixed && defects4j test",
            #    files=(
            #        {
            #            "name": "defects4j.build.xml",
            #            "content": (
            #                _defects4j_root / "defects4j.build.xml"
            #            ).read_bytes(),
            #        },
            #    ),
            # ),
            # Cmd("ls -lha /usr/lib/jvm"),
            # Cmd("ls /usr/lib/jvm/java-8-openjdk-arm64"),
            # Cmd("ls /usr/lib/jvm/java-8-openjdk-amd64"),
            # Cmd("echo $JAVA_HOME"),
            Cmd("cat TESTS-TestSuites.xml"),
        ],
    )


def _get_deepest_path(path: Path, file_only: bool = False) -> Path:
    """
    Returns the deepest path that is a child of the given path.

    If the path is a file, returns the path itself.
    If the path is a directory, returns the deepest path (file or directory) that is a child of the given path,
    depending on the 'file_only' flag.

    Raises  FileNotFoundError if the path does not exist.

    Args:
    ----
        path (Path): The path to search.
        file_only (bool, optional): If True, only consider files in determining the deepest path.
            If False, considers both files and directories. Defaults to False.

    """
    if not path.exists():
        msg = f"The path {path} does not exist."
        raise FileNotFoundError(msg)

    if path.is_file():
        return path

    deepest_path = None
    max_depth = -1

    for current_path in path.rglob("*"):
        if file_only and not current_path.is_file():
            continue  # Skip directories if we're only looking for files

        depth = len(current_path.relative_to(path).parts)
        if depth > max_depth:
            max_depth = depth
            deepest_path = current_path

    if deepest_path is None:
        msg = "No suitable deepest path found under the given directory."
        raise FileNotFoundError(
            msg,
        )

    return deepest_path


def produce_defects4j_problem(
    project_name: str,
    problem_id: str,
    file_name: str,
    line_no: int,
    projects_root: Path | None = None,
) -> CodeProblem:
    projects_root = projects_root or _default_projects_root()

    file_path = Path(f"{project_name}_{problem_id}") / file_name
    source_path = projects_root / file_path
    if not source_path.exists():
        msg = f"Could not find the buggy source file for {source_path}"
        raise Exception(msg)

    source_file = MultilineString(source_path.read_text())

    # source_file[line_no] = "<buggy>" + source_file[line_no].rstrip() + "</buggy>\n"

    known_solution = _figure_out_known_solution_for_defects4j(
        project_name,
        problem_id,
        file_name,
        line_no,
    )

    invisible_spans = []
    print(line_no)

    if line_no > 0:
        invisible_spans.append((0, max(line_no - 10, 1)))

    if line_no < len(source_file) - 1:
        invisible_spans.append(
            (min(len(source_file) - 1, line_no + 10), len(source_file)),
        )

    test_case = make_discovered_test_case(
        project_name,
        problem_id,
        file_name,
        is_buggy=True,
    )

    return make_simple_line_edit_problem(
        code=str(source_file),
        invisible_spans=invisible_spans,
        test_cases=[test_case],
        edit_lines_span=(line_no, line_no + 1),
        current_text_visible=True,
        target_path="solution.java",  # str(Path("/defects4j/Defects4J_projects") / file_path),
        known_solutions=[known_solution],
        lang_spec=JavaLangSpec(),
        dataset_name="defects4j",
        problem_id=problem_id,
        docker_env=DEFECTS_4_J_DOCKER_ENV,
        override_instructions=INSTRUCTIONS_ONE_LINE,
        cap_to_max_same_num_lines=True,
    )


def yield_defects4j(one_line: bool = True) -> Iterator[CodeProblem]:
    warnings.warn(
        "There might be problems with the docker setup currently for defects4j"
    )
    if not one_line:
        raise NotImplementedError
    df = pd.read_csv(
        _defects4j_root / "Defects4J_oneLiner_metadata.csv",
        header=None,
        names=["Project", "ID", "File_Path", "LOC"],
    )

    for _, row in df.iterrows():
        project_name = str(row["Project"])
        problem_id = str(row["ID"])
        file_path = str(row["File_Path"])
        line_no = int(row["LOC"]) - 1

        problem = produce_defects4j_problem(
            project_name=project_name,
            problem_id=problem_id,
            file_name=file_path,
            line_no=line_no,
        )

        if problem:
            yield problem



import json
import zlib
import pickle
import base64
from collections.abc import Iterable
from enum import Enum
from datetime import datetime
from dataclasses import dataclass

from datasets import load_dataset


class Platform(Enum):
    LEETCODE = "leetcode"
    CODEFORCES = "codeforces"
    ATCODER = "atcoder"


class Difficulty(Enum):
    EASY = "easy"
    MEDIUM = "medium"
    HARD = "hard"


class TestType(Enum):
    STDIN = "stdin"
    FUNCTIONAL = "functional"


@dataclass
class Test:
    input: str
    output: str
    testtype: TestType
    func_name: str | None = None

    def __post_init__(self):
        self.testtype = TestType(self.testtype)
        # if self.testtype == TestType.FUNCTIONAL:
        #     self.input = json.loads(self.input)
        #     self.output = json.loads(self.output)


@dataclass
class LivecodeInternalCodeGenerationProblem:
    question_title: str
    question_content: str
    platform: Platform
    question_id: str
    contest_id: str
    contest_date: datetime
    starter_code: str
    difficulty: Difficulty
    public_test_cases: list[Test]
    private_test_cases: list[Test]
    metadata: dict | None

    def __post_init__(self):
        self.platform = Platform(self.platform)
        self.difficulty = Difficulty(self.difficulty)
        self.contest_date = datetime.fromisoformat(self.contest_date)
        if self.metadata is not None:
            self.metadata = json.loads(self.metadata)  # type: ignore
        else:
            self.metadata = {}

        if isinstance(self.public_test_cases, str):
            self.public_test_cases = json.loads(self.public_test_cases)  # type: ignore

        def json_to_tc_list(js):
            func_name = self.metadata.get("func_name", None)
            return [
                Test(
                    **t,
                    func_name=func_name,
                ) if isinstance(t, dict) else t
                for t in js
            ]

        self.public_test_cases = json_to_tc_list(self.public_test_cases)  # type: ignore

        try:
            if isinstance(self.private_test_cases, str):
                self.private_test_cases = json.loads(self.private_test_cases)  # type: ignore
        except:
            self.private_test_cases = json.loads(
                pickle.loads(
                    zlib.decompress(
                        base64.b64decode(self.private_test_cases.encode("utf-8"))  # type: ignore
                    )
                )
            )  # type: ignore
        self.private_test_cases = json_to_tc_list(self.private_test_cases)  # type: ignore


    def insert_output(self, output_list: list[str], code_list: list[str]) -> dict:
        return {
            "question_title": self.question_title,
            "question_content": self.question_content,
            "platform": self.platform.value,
            "question_id": self.question_id,
            "contest_id": self.contest_id,
            "contest_date": self.contest_date.isoformat(),
            "starter_code": self.starter_code,
            "difficulty": self.difficulty.value,
            "output_list": output_list,
            "code_list": code_list,
        }

    def insert_output_evaluation(
        self,
        output_list: list[str],
        code_list: list[str],
        graded_list: list[bool],
        **kwargs,
    ) -> dict:
        output = self.insert_output(output_list, code_list)
        output["graded_list"] = graded_list
        output["pass@1"] = graded_list.count(True) / len(graded_list)
        for k, v in kwargs.items():
            output[k] = v
        return output

    def get_evaluation_sample(self):
        return {
            "input_output": json.dumps(
                {
                    "inputs": [
                        t.input
                        for t in self.public_test_cases + self.private_test_cases
                    ],
                    "outputs": [
                        t.output
                        for t in self.public_test_cases + self.private_test_cases
                    ],
                    "fn_name": self.metadata.get("func_name", None),
                }
            ),
        }


def load_code_generation_dataset(release_version="release_v1") -> Iterable[LivecodeInternalCodeGenerationProblem]:
    dataset = load_dataset(
        "livecodebench/code_generation_lite",
        split="test", version_tag=release_version,
        trust_remote_code=True,
        streaming=True,
    )
    yield from (LivecodeInternalCodeGenerationProblem(**p) for p in dataset)  # type: ignore


def load_code_generation_dataset_not_fast(release_version="release_v1") -> list[LivecodeInternalCodeGenerationProblem]:
    dataset = load_dataset("livecodebench/code_generation", split="test")
    dataset = [LivecodeInternalCodeGenerationProblem(**p) for p in dataset]  # type: ignore
    print(f"Loaded {len(dataset)} problems")
    return dataset


if __name__ == "__main__":
    dataset = load_code_generation_dataset()



from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic, TypeVar


@dataclass
class UncertaintyEstimate:
    name_key: str


@dataclass
class UncertaintyPointEstimate(UncertaintyEstimate):
    value: float


@dataclass
class ProbabilityIsCorrect(UncertaintyPointEstimate):
    pass


T = TypeVar("T", bound=UncertaintyEstimate)


class DelayedUncertaintyEstimate(Generic[T]):
    """
    A return from a model for calculating uncertainty.

    It is potentially delayed as if it is computationally
    expensive to estimate the uncertainty the model might
    decide to lazily compute it only if needed.
    """

    def __init__(self, estimate_callback: Callable[[], T]) -> None:
        self._callback = estimate_callback
        self._value = None
        self._computed = False

    def get_estimate(self) -> T:
        if not self._computed:
            self._value = self._callback()
            self._computed = True
            self._callback = None

        return self._value


class PrecomputedUncertaintyEstimate(DelayedUncertaintyEstimate[T], Generic[T]):
    def __init__(self, estimate: T) -> None:
        super().__init__(lambda: estimate)



from synthegrator.util import (
    estimate_best_overlap_span,
    find_after_first_occurrence_of_string,
)
import pytest


def tokenize(val):
    return val.split()


def test_best_overlap_span():
    span = estimate_best_overlap_span("a b c", None, "a b", tokenize)
    assert span == (0, 2)


def test_best_overlap_span2():
    span = estimate_best_overlap_span("a b c", None, "a b c", tokenize)
    assert span == (0, 3)


def test_best_overlap_span3():
    span = estimate_best_overlap_span("a b c d e", None, "b c d", tokenize)
    assert span == (1, 4)


def test_best_overlap_span4():
    span = estimate_best_overlap_span("a b c", None, "x a b c", tokenize)
    assert span == (0, 3)


@pytest.mark.skip(reason="Need to expand implementation to handle this case")
def test_best_overlap_span_weird():
    span = estimate_best_overlap_span(
        "a b c de f g",
        # Make the larger tokens split more
        ["a", "b", "c", "d", "e", "f", "g"],
        "c de f",
        tokenize,
    )
    assert span == (2, 5)


def test_find_after_first_occurrence_basic():
    assert find_after_first_occurrence_of_string(["My", " Name", " Is"], " Name") == (
        2,
        0,
    )


def test_find_after_first_occurrence_basic2():
    assert find_after_first_occurrence_of_string(["My", " Name", " Is"], "Name") == (
        2,
        0,
    )


def test_find_after_first_occurrence_spanning_parts():
    assert find_after_first_occurrence_of_string(
        ["My", " Name", " Is", " Bob"],
        "Name Is",
    ) == (3, 0)


def test_find_after_first_occurrence_within_part():
    assert find_after_first_occurrence_of_string(
        ["My", " Name", " Is Bob"],
        "Name Is",
    ) == (2, 3)


def test_find_after_first_occurrence_not_found():
    assert find_after_first_occurrence_of_string(
        ["My", " Name", " Is Bob"],
        "Random",
    ) == (None, None)


def test_find_after_first_occurrence_empty_string():
    assert find_after_first_occurrence_of_string([], "") == (None, None)


def test_find_after_first_occurrence_empty_search():
    assert find_after_first_occurrence_of_string(["Hello", "World"], "") == (0, 0)


def test_find_after_first_occurrence_only_part_matches():
    assert find_after_first_occurrence_of_string(["Hello", "World"], "World") == (2, 0)


def test_find_after_first_occurrence_split_str_at_start():
    assert find_after_first_occurrence_of_string(["Hello", " World"], "Hello") == (1, 0)



from pathlib import Path

from examples.run_human_eval import run_for_problems_and_serialize
from synthegrator.synthdatasets.dypybench import yield_dypybench

try:
    import torch
except ImportError:
    torch = None


cur_file = Path(__file__).parent.absolute()


def get_dypy_results(
    output_directory: Path = cur_file / "results",
    read_file_if_exists=False,
):
    print("Gathering all problems...")
    problems = list(yield_dypybench())
    return run_for_problems_and_serialize(
        problems,
        "dypybench",
        output_directory=output_directory,
        read_file_if_exists=read_file_if_exists,
        max_limit=200,
    )


if __name__ == "__main__":
    get_dypy_results()



import pytest

from synthegrator.lang_specs.lang_spec import (
    PythonFunctionParser,
)
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec, LsFunctionDefPython
from synthegrator.lang_specs.lang_util import truncate_when_decreases_indent


@pytest.fixture
def example_dypy_method_nodocstring():
    return """def add_task(
        self,
        task: Task,
        queue: None | BaseTaskQueue = None,
        raise_error: bool = False,
    ) -> bool:
        if queue is None:
            queue = self.task_queue
        if task.priority is None or not task.priority_set_explicitly:
            task.priority = self.generate_task_priority()
            task.priority_set_explicitly = False
        else:
            task.priority_set_explicitly = True

        if not task.request.url or not task.request.url.startswith(
            ("http://", "https://", "ftp://", "file://", "feed://")
        ):
            self.collect_runtime_event("task-with-invalid-url", task.request.url)
            msg = "Invalid task URL: %s" % task.request.url
            if raise_error:
                raise SpiderError(msg)
            logger.error(
                "%s\nTraceback:\n%s",
                msg,
                "".join(format_stack()),
            )
            return False
        # TODO: keep original task priority if it was set explicitly
        # WTF the previous comment means?
        queue.put(task, priority=task.priority, schedule_time=task.schedule_time)
        return True"""


def test_get_prefixcomments_only_at_start():
    langspec_py = PythonLangSpec()
    comments = list(
        langspec_py.get_prefix_comments(
            "def foo():\n"
            "    this_isnot()\n"
            "    '''\n"
            "    This is not a docstring\n"
            "    '''\n"
            "    pass\n"
            "def bar():\n"
            "    this_isnot()\n"
            "    '''\n"
            "    This is not another docstring\n"
            "    '''\n"
            "    pass\n"
            "def boo():\n"
            "    '''\n"
            "    This is a docstring\n"
            "    '''\n"
            "    pass\n",
        ),
    )
    assert len(comments) == 1
    assert comments[0].doc_string_str == "'''\n    This is a docstring\n    '''"


def test_treesitter_getdocstring_only_start1(example_dypy_method_nodocstring):
    parser = PythonFunctionParser(example_dypy_method_nodocstring)
    assert parser.docstring_node is None


def test_treesitter_getdocstring_only_start2(example_dypy_method_nodocstring):
    parser = PythonFunctionParser(
        "def foo():\n"
        "    this_isnot()\n"
        "    '''\n"
        "    This is not a docstring\n"
        "    '''\n"
        "    pass\n",
    )
    assert parser.docstring_node is None


def test_treesitter_getdocstring_only_start3(example_dypy_method_nodocstring):
    parser = PythonFunctionParser(
        "def foo():\n    '''\n    This is a docstring\n    '''\n    pass\n",
    )
    assert (
        parser.docstring_node.text.decode() == "'''\n    This is a docstring\n    '''"
    )


def test_treesitter_getdocstring_only_start1():
    fixture1 = (
        "def foo():\n"
        "    this_isnot()\n"
        "    '''\n"
        "    This is a docstring\n"
        "    '''\n"
        "    pass\n"
    )
    parser = PythonFunctionParser(fixture1)
    assert parser.docstring_node is None


def test_treesitter_getdocstring_only_start2():
    fixture1 = "def foo():\n    '''\n    This is a docstring\n    '''\n    pass\n"
    parser = PythonFunctionParser(fixture1)
    assert parser.docstring_node is not None
    assert (
        parser.docstring_node.text.decode() == "'''\n    This is a docstring\n    '''"
    )


def test_get_docstrings():
    langspec_py = PythonLangSpec()
    comments = list(
        langspec_py.get_prefix_comments(
            "def foo():\n"
            "    '''\n"
            "    This is a docstring\n"
            "    '''\n"
            "    pass\n"
            "def bar():\n"
            "    '''\n"
            "    This is other docstring\n"
            "    '''\n"
            "    pass\n",
        ),
    )
    assert len(comments) == 2
    assert comments[0].doc_string_node.start_point == (1, 4)
    assert comments[0].doc_string_node.end_point == (3, 4 + 3)
    assert (
        comments[0].doc_string_node.text.decode("utf-8")
        == "'''\n    This is a docstring\n    '''"
    )
    assert comments[0].function_node.start_point == (0, 0)
    assert comments[0].function_node.end_point == (4, 4 + 4)


def test_truncate_when_decrease_indent():
    result = truncate_when_decreases_indent("    return 2\nassert False")
    assert result == "    return 2"


def test_truncate_when_decrease_indent2():
    result = truncate_when_decreases_indent(  # fmt: off
        "    x = 2\n"
        "    y = x * 2\n"
        "    for foo in bar:\n"
        "        y += 1\n"
        "    return y\n"
        "assert False",
    )
    assert (
        result
        == "    x = 2\n    y = x * 2\n    for foo in bar:\n        y += 1\n    return y"
    )  # fmt: on


def test_get_body_src_include_prefix_comment_triple_double_quotes():
    """Test get_body_src with include_prefix_comment parameter using triple double quotes"""
    source = '''def foo():
    """This is a docstring"""
    x = 1
    return x'''
    
    langspec = PythonLangSpec()
    tree = langspec.get_tree_sitter_tree(source)
    func_defs = list(PythonFunctionParser.get_function_definitions(tree.root_node))
    func_def = LsFunctionDefPython(func_defs[0], tree)
    
    # With prefix comment (default behavior)
    with_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=True)
    assert with_prefix == '    """This is a docstring"""\n    x = 1\n    return x'
    
    # Without prefix comment
    without_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=False)
    assert without_prefix == '    x = 1\n    return x'


def test_get_body_src_include_prefix_comment_triple_single_quotes():
    """Test get_body_src with include_prefix_comment parameter using triple single quotes"""
    source = '''def bar():
    \'\'\'This is a docstring\'\'\'
    x = 1
    return x'''
    
    langspec = PythonLangSpec()
    tree = langspec.get_tree_sitter_tree(source)
    func_defs = list(PythonFunctionParser.get_function_definitions(tree.root_node))
    func_def = LsFunctionDefPython(func_defs[0], tree)
    
    # With prefix comment (default behavior)
    with_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=True)
    assert with_prefix == "    '''This is a docstring'''\n    x = 1\n    return x"
    
    # Without prefix comment
    without_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=False)
    assert without_prefix == '    x = 1\n    return x'


def test_get_body_src_include_prefix_comment_multiline_docstring():
    """Test get_body_src with multiline docstrings using both quote styles"""
    # Multi-line triple double quotes
    source_double = '''def baz():
    """
    This is a 
    multi-line docstring
    """
    x = 1
    return x'''
    
    # Multi-line triple single quotes
    source_single = '''def qux():
    \'\'\'
    This is a 
    multi-line docstring
    \'\'\'
    x = 1
    return x'''
    
    langspec = PythonLangSpec()
    
    for source in [source_double, source_single]:
        tree = langspec.get_tree_sitter_tree(source)
        func_defs = list(PythonFunctionParser.get_function_definitions(tree.root_node))
        func_def = LsFunctionDefPython(func_defs[0], tree)
        
        # With prefix comment should include docstring
        with_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=True)
        assert 'multi-line docstring' in with_prefix
        assert 'x = 1' in with_prefix
        
        # Without prefix comment should exclude docstring
        without_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=False)
        assert 'multi-line docstring' not in without_prefix
        assert without_prefix == '    x = 1\n    return x'


def test_get_body_src_include_prefix_comment_no_docstring():
    """Test get_body_src with include_prefix_comment when there's no docstring"""
    source = '''def nope():
    x = 1
    return x'''
    
    langspec = PythonLangSpec()
    tree = langspec.get_tree_sitter_tree(source)
    func_defs = list(PythonFunctionParser.get_function_definitions(tree.root_node))
    func_def = LsFunctionDefPython(func_defs[0], tree)
    
    # Both should return the same result when there's no docstring
    with_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=True)
    without_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=False)
    
    assert with_prefix == without_prefix == '    x = 1\n    return x'


def test_get_body_src_include_prefix_comment_invalid_docstring():
    """Test get_body_src with include_prefix_comment when docstring is not at start"""
    source = '''def invalid():
    x = 1
    """This is not a docstring"""
    return x'''
    
    langspec = PythonLangSpec()
    tree = langspec.get_tree_sitter_tree(source)
    func_defs = list(PythonFunctionParser.get_function_definitions(tree.root_node))
    func_def = LsFunctionDefPython(func_defs[0], tree)
    
    # Both should return the same result when there's no valid prefix docstring
    with_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=True)
    without_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=False)
    
    assert with_prefix == without_prefix == '    x = 1\n    """This is not a docstring"""\n    return x'


def test_get_body_src_include_prefix_comment_only_docstring():
    """Test get_body_src with include_prefix_comment when function has only a docstring"""
    source = '''def empty():
    """Just a docstring"""'''
    
    langspec = PythonLangSpec()
    tree = langspec.get_tree_sitter_tree(source)
    func_defs = list(PythonFunctionParser.get_function_definitions(tree.root_node))
    func_def = LsFunctionDefPython(func_defs[0], tree)
    
    # With prefix comment should include docstring
    with_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=True)
    assert with_prefix == '    """Just a docstring"""'
    
    # Without prefix comment should be empty
    without_prefix = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=False)
    assert without_prefix == ''


def test_get_body_src_include_prefix_comment_with_other_params():
    """Test get_body_src with include_prefix_comment combined with other parameters"""
    source = '''def foo():
    """This is a docstring"""
    x = 1
    return x'''
    
    langspec = PythonLangSpec()
    tree = langspec.get_tree_sitter_tree(source)
    func_defs = list(PythonFunctionParser.get_function_definitions(tree.root_node))
    func_def = LsFunctionDefPython(func_defs[0], tree)
    
    # Test with trailing newline
    with_trailing = func_def.get_body_src(include_trailing_new_line=True, include_prefix_comment=False)
    assert with_trailing == '    x = 1\n    return x\n'
    
    # Test without trailing newline
    without_trailing = func_def.get_body_src(include_trailing_new_line=False, include_prefix_comment=False)
    assert without_trailing == '    x = 1\n    return x'


def test_get_body_src_include_prefix_comment_nested_function():
    """Test get_body_src with include_prefix_comment for nested functions"""
    source = '''def outer():
    """Outer function docstring"""
    x = 1
    
    def inner():
        """Inner function docstring"""
        y = 2
        return y
    
    return inner() + x'''
    
    langspec = PythonLangSpec()
    tree = langspec.get_tree_sitter_tree(source)
    func_defs = list(PythonFunctionParser.get_function_definitions(tree.root_node))
    
    # Test outer function
    outer_func = LsFunctionDefPython(func_defs[0], tree)
    
    # With prefix comment should include outer docstring
    with_prefix = outer_func.get_body_src(include_trailing_new_line=False, include_prefix_comment=True)
    assert '"""Outer function docstring"""' in with_prefix
    assert 'def inner():' in with_prefix
    assert '"""Inner function docstring"""' in with_prefix
    
    # Without prefix comment should exclude outer docstring but keep inner
    without_prefix = outer_func.get_body_src(include_trailing_new_line=False, include_prefix_comment=False)
    assert '"""Outer function docstring"""' not in without_prefix
    assert 'def inner():' in without_prefix
    assert '"""Inner function docstring"""' in without_prefix  # Inner docstring should remain
    assert 'x = 1' in without_prefix
    assert 'return inner() + x' in without_prefix



from pathlib import Path
from datasets import load_dataset
from collections import defaultdict
import numpy as np
import json
import pandas as pd
from synthegrator.execution_threading import solve_and_evaluate_problems


cur_path = Path(__file__).parent

known_fails = ['repocod_astropy/utils/data_info.py_File-level_dtype_info_name', 'repocod_xarray/coding/times.py_Repository-level_decode_cf_datetime', 'repocod_sklearn/utils/validation.py_Repository-level_check_X_y', 'repocod_src/datasets/dataset_dict.py_Repository-level_DatasetDict.push_to_hub', 'repocod_src/datasets/features/features.py_Self-contained_Features.reorder_fields_as', 'repocod_sklearn/utils/graph.py_Repository-level__fix_connected_components', 'repocod_sphinx/ext/autodoc/__init__.py_Repository-level_filter_members', 'repocod_sympy/core/sorting.py_Repository-level_default_sort_key', 'repocod_sklearn/model_selection/_validation.py_Repository-level__fit_and_score', 'repocod_sklearn/utils/metaestimators.py_Repository-level__safe_split']


def make_gt_gen(only_known_fails = False):
    from pprint import pprint

    from collections import defaultdict
    from datasets import load_dataset
    #from authorcode.inference.inference_utils import get_problem_instance, reset_instance
    import os
    import json
    from itertools import islice

    data = load_dataset('lt-asset/REPOCOD')['test']
    json_out = defaultdict(dict)
    num_added = 0
    for i, sample in enumerate(data):
        func_name = sample['function_name']
        if only_known_fails and not any(func_name in fail for fail in known_fails):
            continue
        print(sample.keys())
        print(sample['function_name'])
        print(sample['content_class'])
        print(sample['target_module_path'])
        print(sample['repo_id'])
        prompt = sample['prompt']
        full_function = sample['full_function']
        repository = sample['repository']
        assert full_function.startswith(prompt)
        body = full_function[len(prompt):]
        json_out[repository][sample['repo_id']] = {
            'function_name': sample['function_name'],
            'prompt': prompt,
            'output': body
        }
        num_added += 1
    fn = "gt"
    if only_known_fails:
        fn += "_fail"
    fn += f"_{num_added}"
    fn += ".json"

    (cur_path / fn).write_text(json.dumps(json_out, indent=2))


def _make_author_style_from_df():
    from _quick_inspect_df import _make_df
    from lmwrapper.openai_wrapper import get_open_ai_lm, OpenAiModelNames
    df, key = _make_df(1000, OpenAiModelNames.gpt_4o_mini, "autoregressive")

    data = load_dataset('lt-asset/REPOCOD')['test']
    problem_id_to_repo_id = {}
    for sample in data:
        problem_id=f"repocod_{sample['target_module_path']}_{sample['content_class']}_{sample['function_name']}"
        problem_id_to_repo_id[problem_id] = (sample['repo_id'], sample['function_name'], sample['repository'])


    #df = df.sample(20)
    out_json = defaultdict(dict)
    num_added = 0
    for i, row in df.iterrows():
        val = row['solution__solve_steps'][0]['value']
        is_success = row['main_metric__is_success']
        #if not is_success:
        #    continue
        problem_id = row['solution__problem__problem_id']
        repo_id, function_name, repo_name = problem_id_to_repo_id[problem_id]
        out_json[repo_name][repo_id] = {
            'function_name': function_name,
            'prompt': row['solution__prompt__text'][0]['content'],
            'output': val,
        }
        num_added += 1
    print(f"Added {num_added} problems")
    (cur_path / f"mini_synth_{num_added}.json").write_text(json.dumps(out_json, indent=2))




def retry_known_fails_synth():
    from synthegrator.synthdatasets.repocod import yield_repocod_problems
    from synthegrator.code_solver import DummyCodeSolverAutoRegressive
    from lmwrapper.openai_wrapper import get_open_ai_lm, OpenAiModelNames
    problems = []
    for prob in yield_repocod_problems():
        if not any(prob.problem_id in fail for fail in known_fails):
            continue
        print(prob.problem_id)
        problems.append(prob)
    print(len(problems))
    print(problems[0])
    lm = get_open_ai_lm(OpenAiModelNames.gpt_4_1_mini)
    solver = DummyCodeSolverAutoRegressive(lm)
    evls = solve_and_evaluate_problems(solver, problems, max_threads_eval=1, max_threads_solve=1)
    from tqdm import tqdm
    for evl in tqdm(evls):
        print(evl.solution.problem.problem_id)
        print(evl.main_metric.is_success)
    exit()


def calc_results_for_author_style(
    name: str,
):
    data = load_dataset('lt-asset/REPOCOD')['test']
    repo_name_and_id_to_problem_id = {}
    for sample in data:
        problem_id=f"repocod_{sample['target_module_path']}_{sample['content_class']}_{sample['function_name']}"
        repo_name_and_id_to_problem_id[(sample['repository'], sample['repo_id'])] = problem_id

    path = cur_path / f"authorcode/evaluate/evaluate_results/{name}.json"
    data = json.loads(path.read_text())
    all_results = []
    for repo_name, repo_data in data.items():
        print(repo_name)
        for repo_id, prob_data in repo_data.items():
            all_results.append({
                "repo_name": repo_name,
                "repo_id": repo_id,
                "problem_id": repo_name_and_id_to_problem_id[(repo_name, repo_id)],
                **prob_data,
            })
    print(all_results)
    df = pd.DataFrame(all_results)
    print(df.groupby("repo_name")["result"].mean())
    print(df['result'].mean())
    return df


if __name__ == "__main__":
    #make_gt_gen()
    #retry_known_fails_synth()
    #_make_author_style_from_df()
    print("GT")
    #calc_results_for_author_style("gt_980")
    print("-"*100)
    print("Mini synth")
    calc_results_for_author_style("mini_synth_980")


"""
Utilities for serializing results to DataFrames
for analysis, visualization, and persistence.
"""

import gzip
import pickle
import random
from collections.abc import Callable, Iterable

import pandas as pd
from lmwrapper.utils import flatten_dict

from synthegrator.code_problems import CodeProblem
from synthegrator.solution_eval import (
    SolutionEvaluation,
    SolutionMetric,
    TestRunResultList,
)


def problem_to_dict(
    code_problem: CodeProblem,
    encode_transformation_spec: bool = False,
    encode_wd: bool = False,
    include_pickle_gzip: bool = False,
) -> dict:
    """
    Converts a CodeProblem to a dict, for serialization.

    Args:
    ----
        code_problem: The CodeProblem to convert.
        encode_transformation_spec: NOT IMPLEMENTED
        encode_wd: NOT IMPLEMENTED
        include_pickle_gzip: If True, includes a gzip-compressed pickle of the
            CodeProblem in the output dict. This is mostly a stopgap until
            have more complete serialization and deserialization.

    """
    if encode_transformation_spec:
        raise NotImplementedError
    if encode_wd:
        raise NotImplementedError
    out = {
        "problem_id": code_problem.problem_id,
        "dataset_name": code_problem.dataset_name,
    }
    if include_pickle_gzip:
        try:
            out["pickle_gzip"] = gzip.compress(pickle.dumps(code_problem), mtime=0)
        except TypeError as e:
            out["pickle_gzip"] = None
    return out


def _dict_key_prepend(prefix: str, d: dict) -> dict:
    if d is None:
        return {}
    return {prefix + "__" + k: v for k, v in d.items()}


def _prop_to_dict(obj, prop, serializer) -> dict:
    if serializer is None:
        return {}
    return _dict_key_prepend(prop, serializer(getattr(obj, prop)))


def test_results_list_to_dict(test_results_list: TestRunResultList) -> dict:
    prefix = "test_results_list"
    if test_results_list is None:
        return {}
    return {
        f"{prefix}__runtime": test_results_list.runtime,
        f"{prefix}__syntax_check_result": test_results_list.syntax_check_result,
        f"{prefix}__len": len(test_results_list),
        f"{prefix}__collection_error": test_results_list.collection_error,
        f"{prefix}__exec_error": test_results_list.exec_error,
        f"{prefix}__timeout": test_results_list.timeout,
    }


def solution_eval_to_dict(
    solution_eval: SolutionEvaluation,
    test_results_list_to_dict: Callable[[TestRunResultList], dict] | None = test_results_list_to_dict,
    pickle_gzip_whole_solution_eval: bool = False,
) -> dict:
    out = {
        "solution": solution_eval.solution.dict_serialize(),
        **solution_metric_to_dict(solution_eval.main_metric, is_main=True),
    }
    for metric in solution_eval.extra_metrics:
        out.update(solution_metric_to_dict(metric, is_main=False))
    if test_results_list_to_dict is not None:
        out.update(test_results_list_to_dict(solution_eval.test_results))
    if solution_eval.exception is not None:
        out["exception"] = str(solution_eval.exception)
    if pickle_gzip_whole_solution_eval:
        out["pickle_gzip"] = gzip.compress(pickle.dumps(solution_eval), mtime=0)
    return out


def solution_metric_to_dict(
    metric: SolutionMetric,
    is_main: bool,
) -> dict:
    if metric is None:
        return {}
    key = "main_metric" if is_main else metric.label
    out = {
        f"{key}__is_success": metric.is_success,
        f"{key}__float_val": metric.float_val,
    }
    if is_main:
        out[f"{key}__value"] = float(
            metric.is_success if metric.is_success is not None else metric.float_val,
        )
    return out


def solution_evals_to_df(
    solve_evals: Iterable[SolutionEvaluation],
    solver_key: str | None = None,
    solution_eval_to_dict: Callable[[SolutionEvaluation], dict] = solution_eval_to_dict,
    pickle_gzip_whole_solution_eval=False,
) -> pd.DataFrame:
    data = []
    if solver_key is None:
        solver_key = f"unspecified-{hex(random.randint(0, 1000000))}"
    for eval in solve_evals:
        serial = flatten_dict(
            solution_eval_to_dict(
                eval,
                pickle_gzip_whole_solution_eval=pickle_gzip_whole_solution_eval,
            ),
            verify_keys_do_not_have_sep=False,
        )
        serial["solver_key"] = solver_key
        data.append(serial)
    return pd.DataFrame(data)



import dataclasses

from lmwrapper.abstract_predictor import LmPredictor
from lmwrapper.structs import ChatGptRoles, LmChatDialog, LmChatTurn

from synthegrator.code_problems import CodeProblem
from synthegrator.few_shotting import FewShotConfig
from synthegrator.problem_rendering import LmPromptRender, PromptRenderer
from synthegrator.response_parser import (
    AnswerForSolutionInvalidForSolve,
    AnswerForSolutionNeverSupported,
    ResponseParser,
)


def get_answer_for_each(
    problems: list[CodeProblem],
    renderer: PromptRenderer,
    parser: ResponseParser,
    lm: LmPredictor,
) -> list[tuple[LmPromptRender, str]]:
    out = []
    for problem in problems:
        render = renderer.render(problem, lm)
        if len(problem.known_solutions) == 0:
            continue
        try:
            ans = parser.answer_for_solution(
                render, problem, problem.known_solutions[0].solve_steps
            )
        except AnswerForSolutionNeverSupported:
            return []
        except AnswerForSolutionInvalidForSolve:
            continue
        ans_starts_with_exactly_one_space = ans.startswith(" ") and not ans.startswith(
            "  "
        )
        if lm.is_chat_model and ans_starts_with_exactly_one_space:
            ans = ans[1:]
        out.append((render, ans))
    return out


def get_default_text_seperator() -> str:
    return "\n###\n"


def combine_for_few_shot(
    render_and_answers: list[tuple[LmPromptRender, str]],
    lm: LmPredictor,
    text_seperator: str = get_default_text_seperator(),
) -> LmChatDialog | str:
    if lm.is_chat_model:
        messages = []
        for render, ans in render_and_answers:
            if not render.prompt.is_text_a_chat():
                text = render.prompt.get_text_as_chat()
            else:
                text = render.prompt.text
            assert text[-1].role == ChatGptRoles.user
            messages.extend(text)
            messages.append(LmChatTurn(ChatGptRoles.assistant, ans))
        return LmChatDialog(messages)
    messages = []
    for render, ans in render_and_answers:
        assert not render.prompt.is_text_a_chat()
        text = render.prompt.text
        assert isinstance(text, str)
        messages.append(text + ans)
    if len(messages) == 0:
        return ""
    return text_seperator.join(messages) + text_seperator


def generic_prepare_few_shot(
    target_problem: CodeProblem,
    few_shot_config: FewShotConfig,
    renderer: PromptRenderer,
    parser: ResponseParser,
    lm: LmPredictor,
) -> LmPromptRender:
    problems = few_shot_config.library.sample_problems(
        n=few_shot_config.num_examples,
        target_problem=target_problem,
    )
    render_and_answers = get_answer_for_each(problems, renderer, parser, lm)
    few_shot_context = combine_for_few_shot(render_and_answers, lm)
    target_render = renderer.render(target_problem, lm)
    new_prompt = dataclasses.replace(
        target_render.prompt,
        text=few_shot_context
        + (
            target_render.prompt.get_text_as_chat()
            if lm.is_chat_model
            else target_render.prompt.get_text_as_string_default_form()
        ),
    )
    new_render = dataclasses.replace(
        target_render,
        prompt=new_prompt,
    )
    return new_render


class FewShotRendererWrapper(PromptRenderer):
    """
    Wraps a prompt render to be add support
    for few-shotting for a given renderer as long
    as the response parser supports answer_for_solution.
    """

    def __init__(
        self,
        renderer: PromptRenderer,
        response_parser: ResponseParser,
        few_shot_config: FewShotConfig,
    ):
        super().__init__()
        self.renderer = renderer
        self.parser = response_parser
        self.few_shot_config = few_shot_config

    def render(
        self, problem: CodeProblem, lm: LmPredictor, prompt_seed: int | None = None
    ) -> LmPromptRender:
        return generic_prepare_few_shot(
            problem, self.few_shot_config, self.renderer, self.parser, lm
        )



from unittest.mock import MagicMock

import pytest
from lmwrapper.structs import ChatGptRoles

from synthegrator.code_problem_builders import make_simple_method_completion_problem
from synthegrator.code_problems import (
    TestCaseMethodCallIsEq,
)
from synthegrator.few_shotting import FewShotConfig, FewShotLibrary
from synthegrator.problem_rendering import (
    render_markup_to_text_autoregressive,
    render_single_edit_codeproblem_dialog,
)
from synthegrator.prompting_test_case_selection import (
    RandomVisibleTestCaseSelectionStrategy,
)
from synthegrator.tests.non_dataset_tests.test_java_stuff import get_java_double_problem
from synthegrator.transformation_spec import (
    MarkElement,
    MarkText,
    StsInsert,
    StsInvisible,
    StsNormalText,
)


def test_markup_autoregressive_rendering():
    markup = MarkElement(
        verb=StsNormalText(),
        children=[
            MarkText("foo", 0, 3),
        ],
        start_char=0,
        end_char=3,
    )
    problem = MagicMock()
    problem.instructions_are_essential = False
    out = render_markup_to_text_autoregressive(
        markup,
        path="solution.py",
        problem=problem,
    )
    assert out.prompt.text == "foo"
    assert out.path == "solution.py"
    assert out.changed_element is None


def test_markup_autoregressive_rendering2():
    markup = MarkElement(
        verb=StsNormalText(),
        children=[
            MarkText("foo", 0, 3),
            MarkElement(StsInsert(), [], 3, 3),
            MarkText("bar", 3, 6),
        ],
        start_char=0,
        end_char=6,
    )
    problem = MagicMock()
    problem.instructions_are_essential = False
    out = render_markup_to_text_autoregressive(
        markup,
        path="solution.py",
        problem=problem,
    )
    assert out.prompt.text == "foo"
    assert out.path == "solution.py"
    assert out.changed_element == markup.children[1]


def test_markup_autoregressive_rendering3():
    markup = MarkElement(
        verb=StsNormalText(),
        children=[
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("hide", 3, 7)], 3, 7),
            MarkText("bar", 7, 10),
        ],
        start_char=0,
        end_char=10,
    )
    problem = MagicMock()
    problem.instructions_are_essential = False
    out = render_markup_to_text_autoregressive(
        markup,
        path="solution.py",
        problem=problem,
    )
    assert out.prompt.text == "foobar"
    assert out.path == "solution.py"
    assert out.changed_element is None


def make_test_case(test_input, output, method_name):
    return TestCaseMethodCallIsEq(
        test_input=test_input,
        output=output,
        method_name=method_name,
        fail_message="",
        is_hidden_test=False,
    )


@pytest.fixture
def factor_problems():
    return [
        make_simple_method_completion_problem(
            signature="def double(x) -> int:",
            prompt="Write a function that doubles its input.",
            known_solutions=[
                "    return x * 2\n",
            ],
            test_cases=[
                make_test_case((2,), 4, "double"),
                make_test_case((3,), 6, "double"),
            ],
            target_path="foo.py",
            override_instructions="Complete the following method please",
            problem_id="double",
        ),
        make_simple_method_completion_problem(
            signature="def triple(x) -> int:",
            prompt="Write a function that triples its input.",
            known_solutions=[
                "    return x * 3\n",
            ],
            test_cases=[
                make_test_case((2,), 6, "triple"),
                make_test_case((3,), 9, "triple"),
            ],
            target_path="foo.py",
            override_instructions="Complete the following method please",
            problem_id="triple",
        ),
        make_simple_method_completion_problem(
            signature="def quadruple(x) -> int:",
            prompt="Write a function that quadruples its input.",
            known_solutions=[
                "    return x * 4\n",
            ],
            test_cases=[
                make_test_case((2,), 8, "quadruple"),
                make_test_case((3,), 12, "quadruple"),
            ],
            target_path="foo.py",
            override_instructions="Complete the following method please",
            problem_id="quadruple",
        ),
    ]


@pytest.fixture
def factor_problems_java():
    return [get_java_double_problem()]


def test_dialog_rendering(factor_problems):
    problem = factor_problems[0]
    render = render_single_edit_codeproblem_dialog(problem)
    assert render.path == "foo.py"
    assert len(render.prompt.text) == 1
    content = (
        "Complete the following method please\n"
        "```\n"
        "def double(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n'
        "```\n"
    )
    assert render.prompt.text[0].content == content
    assert render.prompt.text[0].role == ChatGptRoles.user
    element = render.changed_element
    assert isinstance(element, MarkElement)
    assert element.verb == StsInsert(
        set_base_space_indent=4,
        stop_at_block_end=True,
    )
    assert element.mark_end_original_char == len(
        "def double(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n',
    )


def test_dialog_rendering_few_shot(factor_problems):
    problem = factor_problems[0]
    render = render_single_edit_codeproblem_dialog(
        problem,
        few_shot_config=FewShotConfig(
            library=FewShotLibrary(factor_problems[:2]),
            num_examples=1,
            change_examples_between_solves=False,
        ),
        prompt_seed=0,
    )
    assert render.path == "foo.py"
    assert len(render.prompt.text) == 3
    content = (
        "Complete the following method please\n"
        "```\n"
        "def double(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n'
        "```\n"
    )
    assert render.prompt.text[2].content == content
    element = render.changed_element
    assert isinstance(element, MarkElement)
    assert element.verb == StsInsert(
        set_base_space_indent=4,
        stop_at_block_end=True,
    )
    assert element.mark_end_original_char == len(
        "def double(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n',
    )
    assert (
        render.prompt.text[0].content == "Complete the following method please\n"
        "```\n"
        "def triple(x) -> int:\n"
        '    """\n'
        "    Write a function that triples its input.\n"
        '    """\n'
        "```\n"
    )
    assert render.prompt.text[1].content == "```\n    return x * 3\n```"
    assert render.prompt.text[0].role == ChatGptRoles.user
    assert render.prompt.text[1].role == ChatGptRoles.assistant
    assert render.prompt.text[2].role == ChatGptRoles.user


def test_dialog_rendering_few_shot_test_case_prompt(factor_problems):
    problem = factor_problems[0]
    render = render_single_edit_codeproblem_dialog(
        problem,
        few_shot_config=FewShotConfig(
            library=FewShotLibrary(factor_problems[:2]),
            num_examples=1,
            change_examples_between_solves=False,
        ),
        prompt_seed=0,
        prompt_test_case_selection_strategy=RandomVisibleTestCaseSelectionStrategy(0),
    )
    assert render.path == "foo.py"
    assert len(render.prompt.text) == 3
    content = (
        "Complete the following method please\n"
        "```\n"
        "def double(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n'
        "```\n"
        "Test cases:\n"
        "```\n"
        "def test_double_0():\n"
        "    assert double(2) == 4\n"
        "def test_double_1():\n"
        "    assert double(3) == 6\n"
        "```"
    )
    assert render.prompt.text[2].content == content
    element = render.changed_element
    assert isinstance(element, MarkElement)
    assert element.verb == StsInsert(
        set_base_space_indent=4,
        stop_at_block_end=True,
    )
    assert element.mark_end_original_char == len(
        "def double(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n',
    )
    assert (
        render.prompt.text[0].content == "Complete the following method please\n"
        "```\n"
        "def triple(x) -> int:\n"
        '    """\n'
        "    Write a function that triples its input.\n"
        '    """\n'
        "```\n"
        "Test cases:\n"
        "```\n"
        "def test_triple_0():\n"
        "    assert triple(2) == 6\n"
        "def test_triple_1():\n"
        "    assert triple(3) == 9\n"
        "```"
    )
    assert render.prompt.text[1].content == "```\n    return x * 3\n```"
    assert render.prompt.text[0].role == ChatGptRoles.user
    assert render.prompt.text[1].role == ChatGptRoles.assistant
    assert render.prompt.text[2].role == ChatGptRoles.user



from pathlib import Path

from examples.run_human_eval import (
    get_experiment_solver_constructors,
    run_for_problems_and_serialize,
)
from synthegrator.synthdatasets.defects4j import yield_defects4j

cur_file = Path(__file__).parent.absolute()


def get_defects4j_one_line_results(
    output_directory: Path = cur_file / "results",
    read_file_if_exists=False,
):
    print("Gathering all problems...")
    problems = list(yield_defects4j())
    return run_for_problems_and_serialize(
        problems,
        "defects4j_one_line",
        output_directory=output_directory,
        read_file_if_exists=read_file_if_exists,
        solver_constructors=get_experiment_solver_constructors(use_edit_renders=True),
    )


if __name__ == "__main__":
    get_defects4j_one_line_results()



import random
from abc import ABC, abstractmethod

from synthegrator.code_problems import CodeProblem, TestCase


class PromptingTestCaseSelectionStrategy(ABC):
    """
    A strategy for selecting test cases to help the solver solve a problem.
    This is particularly necessary when doing programming-by-example, where
    the test cases are the main way the solver is specified the problem
    """

    @abstractmethod
    def select_for_problem(
        self,
        problem: CodeProblem,
        max_num_test_cases: int,
    ) -> list[TestCase]:
        raise NotImplementedError


class RandomVisibleTestCaseSelectionStrategy(
    PromptingTestCaseSelectionStrategy,
):
    def __init__(
        self,
        seed: int | None = None,
        cap_test_cases: int | None = None,
        randomize_order_even_when_less_than_cap: bool = False,
    ):
        self._seed = seed
        self._cap_test_cases = cap_test_cases
        self._randomize_order_even_when_less_than_cap = (
            randomize_order_even_when_less_than_cap
        )

    def select_for_problem(
        self,
        problem: CodeProblem,
        max_num_test_cases: int | None = None,
    ) -> list[TestCase]:
        if max_num_test_cases is None:
            max_num_test_cases = len(problem.test_cases)
        if self._cap_test_cases is not None:
            max_num_test_cases = min(max_num_test_cases, self._cap_test_cases)
        if max_num_test_cases <= 0:
            msg = "Max_num_test_cases must be greater than 0"
            raise ValueError(msg)

        available_test_cases = problem.get_non_hidden_test_cases()
        if (
            len(available_test_cases) <= max_num_test_cases
            and not self._randomize_order_even_when_less_than_cap
        ):
            return available_test_cases
        r = random.Random(
            f"{problem.problem_id}{problem.dataset_name}{self._seed}",
        )
        return r.choices(
            available_test_cases,
            k=min(max_num_test_cases, len(available_test_cases)),
        )



from pathlib import Path

import pandas as pd
import concurrent.futures

from examples.run_defects4j import get_defects4j_one_line_results
from examples.run_dypy import get_dypy_results
from examples.run_human_eval import get_human_eval_results
from examples.run_mbpp import get_mbpp_results
from lmwrapper.openai_wrapper import OpenAiModelNames
cur_path = Path(__file__).parent.absolute()
output_dir = cur_path / "results"
output_dir.mkdir(exist_ok=True)


def run_everything():
    print("Gather human eval results...")
    human_eval_results = get_human_eval_results(
        output_directory=output_dir,
        read_file_if_exists=True,
    )
    print(human_eval_results.columns)

    print("Gather mbpp results...")
    mbpp_results = get_mbpp_results(
        output_directory=output_dir,
    )
    print("Gather defects4j results...")
    defects4j_results = get_defects4j_one_line_results(
        output_directory=output_dir,
    )
    print("Gather dypybench results...")
    dypy_results = get_dypy_results(
        output_directory=output_dir,
    )
    print("Done")
    return pd.concat(
        [
            human_eval_results,
            mbpp_results,
            defects4j_results,
            dypy_results,
        ]
    )


def read_all_parquets():
    def read_parquet_file(file):
        return pd.read_parquet(file)
    
    parquet_files = list(output_dir.glob("*.parquet"))
    with concurrent.futures.ThreadPoolExecutor() as executor:
        dfs = list(executor.map(read_parquet_file, parquet_files))
    
    return pd.concat(dfs) if dfs else pd.DataFrame()


def make_table(df):
    tasks = ["human_eval", "mbpp_func", "defects4j_one_line", "dypybench"]
    # print(df['solver_key'].unique())
    lines = [
        "\\begin{tabular}{ll|" + "r" * len(tasks) + "}",
        "LM Name & Solver & HumanEval & MBPP-f & Defects4J-1L & DyPyBench-LC \\\\ \\hline",
    ]
    for lm_name_friendly, solver_key, solver_type_friendly in [
        ("Codegen-350M", "Salesforce/codegen-350M-multi", "Completion"),
        ("Mistral-7B", "mistralai/Mistral-7B-v0.1", "Completion"),
        ("QwenCoder 2.5 0.5B", "Qwen/Qwen2.5-Coder-0.5B", "Completion"),
        #("GPT-3.5-Turbo-Instruct", "gpt_3_5_turbo_instruct", "Completion"),
        #("Qwen3-4B", "Qwen/Qwen3-4B", "Dialogue"),
        ("GPT-3.5-Turbo-Instruct", str(OpenAiModelNames.gpt_3_5_turbo_instruct), "Completion"),
        ("GPT-3.5-Turbo", str(OpenAiModelNames.gpt_3_5_turbo), "Dialogue"),
        #("GPT-4-Turbo", "gpt_4_turbo", "Dialogue"),
        ("GPT-4.1-Nano", str(OpenAiModelNames.gpt_4_1_nano), "Dialogue"),
        ("GPT-4.1", str(OpenAiModelNames.gpt_4_1), "Dialogue"),
    ]:
        row = [lm_name_friendly, solver_type_friendly]
        for task_name in tasks:
            filt_df = df[(df.solver_key == solver_key) & (df.task_name == task_name)]
            if len(filt_df) == 0:
                row.append("-")
                continue
            val = filt_df.main_metric__is_success.mean()
            row.append(f"{val * 100:.1f}\\%")
        lines.append(" & ".join(row) + "\\\\")
    lines.append("\\end{tabular}")
    return "\n".join(lines)


def main():
    df = run_everything()
    df = read_all_parquets()
    print(df)
    print(make_table(df))


if __name__ == "__main__":
    main()



import importlib
import pkgutil

import pytest

import examples


def discover_modules(package):
    """Discover all submodules in a package."""
    package_path = package.__path__[0]
    for _, name, is_pkg in pkgutil.iter_modules([package_path]):
        if not is_pkg:
            yield name


@pytest.mark.parametrize("module_name", discover_modules(examples))
def test_dynamic_import(module_name):
    """Dynamically imports modules from 'examples' to catch static/syntax errors."""
    try:
        # Dynamically import the module
        module = importlib.import_module(f"examples.{module_name}")
        assert module  # Simple check to see if the module is not None
    except Exception as e:
        pytest.fail(f"Failed to import {module_name} due to: {e}")


@pytest.mark.parametrize(
    "module",
    [name for _, name, _ in pkgutil.iter_modules(["..examples"])],
)
def test_import_all_from_examples(module):
    """
    Test to dynamically import all modules in 'examples'
    package to catch static/syntax errors.
    """
    package_name = "examples"
    full_module_name = f"{package_name}.{module}"
    try:
        importlib.import_module(full_module_name)
        assert True  # Import succeeded
    except Exception as e:
        pytest.fail(f"Failed to import {full_module_name}: {e}")



"""
Module defines a few concepts.

1. `TransformationSpec`: this is a specification that executes on a
    working directory. It defines what text and files the solver
    has access to and how it is allowed to manipulate the working directory.
    A spec is made up of several "statements" (StsStatement). Each statement contains
    some selectors which narrow down the statement to parts of code,
    and then a verb which is applied on each of those parts of code.
2. "marking up": A marked up version of of a file is an xml-like version
    of a file. It is the result of applying the `TransformationSpec`
    to the file. This represented as a tree of `StsMarkupNode`s. For example a
    simplified example might be
    ```xml
    <StsNormalText>
    def foo(x):
    <StsInsert/>
    </StsNormalText>
    ```
    representing a method completion problem.
3. `markup_path()`: This function will execute a `TransformationSpec` on
    a file and return the marked up version of the file.

The design decouples the contents of a working directory from the specification
from the actual content of the files. An alternative design might not have
had the concept of a `TransformationSpec` and instead just stored a working
directory filled with already marked up files (so basically just make it all
a bunch of xml-like files). I chose the decoupled design for two reasons:

1. We might guess that most files in a project will be untouched and possibly
   reused in multiple problems. Treating the spec as program that executes on a
   working directory allows the code problem to hopefully store minimally unique
   information and multiple code problems can share the same working directory.
2. Might allow more complex specifications (like allowing the solver to generate
   new files, or make queries to some execution or analysis system). These
   might be hard to specify in the "working directory of xml-like files" design.

However, it definitely does add extra complexity. It is not clear how much
point (1) really matters, as expect we will end up eventually caching a lot
of the markedup files in the future to avoid reexecuting the spec each time.
Depending on what we find, we might simplify the design future. However,
this design hopefully gives us a lot of expressivity as needed.
"""

import dataclasses
import json
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Generator, Iterator, Sequence
from dataclasses import dataclass, field
from functools import cached_property
from re import Pattern
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
)
from xml.sax import saxutils

import numpy as np
import xxhash
from typeguard import typechecked

from synthegrator.environments import ProjectWorkingDirectory
from synthegrator.lang_specs.lang_spec import LangSpec, lang_spec_for_path
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.memory_fs import ProjectDir
from synthegrator.util import hash_string_base62, normalize_probs_dict

MARK_KEY_TYPE = str


@typechecked
def markup_path(
    wd: ProjectWorkingDirectory,
    path: str,
    spec: "TransformationSpec",
    encoding: str = "utf-8",
) -> "MarkElement":
    """
    Executes a `TransformationSpec` on a file in a working directory.
    Any selectors that apply to the file will be performed on it
    """
    content = wd.files[path].content
    content = content.decode(encoding)
    node = MarkElement(
        verb=StsNormalText(),
        children=[MarkText(content, 0, len(content))],
        start_char=0,
        end_char=len(content),
    )
    if spec is None:
        return node
    for stmt in spec.statements:
        path_selector = stmt.path_selector
        if (
            path_selector is not None  # None path selectors match everything
            and not path_selector.does_match_path(path)
        ):
            continue
        text_selector = stmt.text_selector
        if text_selector is None:
            spans = [(0, len(content))]
        else:
            spans = text_selector.get_matching_spans(content, path)
        for start, end in spans:
            node.insert_node(stmt.verb, start, end)
    return node.freeze()


@typechecked
def get_mark_element(
    spec: "TransformationSpec",
    path: str,
    wd: ProjectWorkingDirectory,
    mark_id: MARK_KEY_TYPE,
) -> Union["MarkElement", None]:
    marked_up = markup_path(
        wd=wd,
        path=path,
        spec=spec,
    )
    if marked_up is None:
        return None
    for node in marked_up.depth_first_iterate():
        if isinstance(node, MarkElement) and node.mark_id == mark_id:
            return node
    return None


regex_all = re.compile(r".+")


def visible_lines_is_editable(
    markedup_text: "MarkElement",
) -> np.ndarray:
    is_editable_str_parts = []
    # Our approach will be to construct a new line where all
    # the non-new-line spots that are editable are marked with a '1' and
    # all the spots that are not editable are marked with a '0'.
    # We will then split the string on newlines and convert the strings
    # to ints and threshold them to get the final array.
    # This is just an approach to handle some wackyness where
    #   there is a mix of editable and non-editable on a line
    for state in iterate_parser_states_for_node(markedup_text):
        node = state.node
        if not state.is_visible:
            continue
        if isinstance(node, MarkText):
            lines = node.text.splitlines(keepends=True)
            val = "1" if state.is_editable else "0"
            is_editable_str_parts.append(val)
            if len(lines) > 1:
                for _l in lines[1:]:
                    is_editable_str_parts.append("\n")
                    is_editable_str_parts.append(val)
    joined = "".join(is_editable_str_parts)
    return np.array([int(x) > 0 for x in joined.splitlines(keepends=False)])


def get_verbs_per_path(
    wd: ProjectWorkingDirectory,
    spec: "TransformationSpec",
) -> dict[str, list["StsVerb"]]:
    """
    A helper method that returns a mapping from paths to all the verbs
    used in the path
    """
    paths: dict[str, list[StsVerb]] = defaultdict(list)
    files = wd.files
    for stmt in spec:
        if stmt.path_selector is None:
            for path, _node in files.walk(include_dirs=False):
                paths[path].append(stmt.verb)
        else:
            for path in stmt.path_selector.yield_matching_paths(wd.files):
                paths[path].append(stmt.verb)
    return paths


def map_paths_to_path_ids(
    wd: ProjectWorkingDirectory,
    spec: "TransformationSpec",
) -> dict[str, str]:
    all_paths = list(get_verbs_per_path(wd, spec).keys())
    all_paths.sort()
    return {path: hex(i)[2:] for i, path in enumerate(all_paths)}


"""
### TransformationSpec
"""


class TransformationSpec:
    """
    A transformation spec acts like a little program that specifies ways that
    code can be transformed by a solver in order solve a problem.
    """

    def __init__(self, statements: tuple["StsSpecStatement", ...]):
        self._statements = statements
        self._verify_all_static_marks_unique()

    def _verify_all_static_marks_unique(self):
        static_marks = {}
        for stmt in self._statements:
            if stmt.operation_is_static():
                mark_id = stmt.static_mark_id
                if mark_id in static_marks:
                    msg = "Static verbs must currently all be unique"
                    raise ValueError(msg)
                static_marks[mark_id] = True

    def __iter__(self):
        return iter(self.statements)

    @property
    def statements(self):
        return self._statements

    def editing_statements(self) -> Generator["StsSpecStatement", Any, None]:
        for stmt in self.statements:
            if isinstance(stmt.verb, StsPlaceTransforming):
                yield stmt

    def count_editing_statements(self) -> int:
        return len(list(self.editing_statements()))

    def question_statements(self) -> Generator["StsSpecStatement", Any, None]:
        for statement in self.statements:
            if isinstance(statement.verb, StsValueQuery):
                yield statement

    def classification_question_statements(
        self,
    ) -> Generator["StsSpecStatement", Any, None]:
        for statement in self.question_statements():
            if isinstance(statement.verb, StsCodeClassificationQuestion):
                yield statement

    def __eq__(self, other):
        if not isinstance(other, TransformationSpec):
            return False
        return self.statements == other.statements

    def __repr__(self):
        return f"TransformationSpec(statements={self.statements})"


@dataclass(frozen=True)
class StsSpecStatement:
    """One statement in a `TransformationSpec` program"""

    path_selector: Union["StsPathSelector", None, str]
    """Selects paths to apply to the statement too.
    If None, applies to all files (not directories). If it is a string,
    then it will automatically be converted to a `StsPath` selector."""
    text_selector: Union["StsTextSelector", None]
    """Selects text in the file to apply to the statement too. If None, then
    the entire file will be selected."""
    verb: "StsVerb"
    """The verb that gets applied to the selected text"""

    def __post_init__(self):
        if isinstance(self.path_selector, str):
            object.__setattr__(
                self,
                "path_selector",
                StsPath(self.path_selector),
            )

    def attempt_to_get_path_as_exact_path(self):
        if isinstance(self.path_selector, StsPath):
            if self.path_selector.is_exact_path():
                return self.path_selector.path_glob
            msg = "Path selector is not an exact path"
            raise ValueError(msg)
        msg = "unknown handled path"
        raise ValueError(msg)

    def operation_is_static(self) -> bool:
        return self.verb.operation_is_static()

    @property
    def static_mark_id(self) -> str:
        return self.verb.static_mark_id


@dataclass(frozen=True)
class StsSelector:
    pass


@dataclass(frozen=True)
class StsPathSelector(StsSelector, ABC):
    @abstractmethod
    def does_match_path(self, path: str) -> bool:
        raise NotImplementedError

    def yield_matching_paths(self, project_dir: ProjectDir) -> Iterator[str]:
        for path, _node in project_dir.walk():
            if self.does_match_path(path):
                yield path


@dataclass(frozen=True)
class StsPath(StsPathSelector):
    """A path selector that matches on a given glob (recursive blobs not fully implemented)"""

    path_glob: str

    @cached_property
    def regex(self) -> Pattern:
        val = self.path_glob.replace("**", ".*")
        val = val.replace("*", "[^/]*")
        val = val.replace("?", "[^/]")
        val = f"^{val}$"
        return re.compile(val)

    def does_match_path(self, path: str) -> bool:
        """Checks to see if the path matches performing glob expansion"""
        return self.regex.fullmatch(path) is not None

    def is_exact_path(self) -> bool:
        return "*" not in self.path_glob and "?" not in self.path_glob

    def yield_matching_paths(self, project_dir: ProjectDir) -> Iterator[str]:
        yield from project_dir.expand_glob(self.path_glob)


@dataclass(frozen=True)
class StsTextSelector(StsSelector):
    def get_matching_spans(
        self,
        text: str,
        path: str,
    ) -> Sequence[tuple[int, int]]:
        raise NotImplementedError


@dataclass(frozen=True)
class StsSelectorCharSpan(StsTextSelector):
    """Selects a fixed span of characters in the text"""

    start_char: int
    end_char: int

    def get_matching_spans(
        self,
        text: str,
        path: str,
    ) -> Sequence[tuple[int, int]]:
        return [(self.start_char, self.end_char)]


@dataclass(frozen=True)
class StsPrefixCommentSelector(StsTextSelector):
    """Selects the lines corresponding to prefix comments"""

    def get_matching_spans(
        self,
        text: str,
        path: str,
    ) -> Sequence[tuple[int, int]]:
        lang_spec = lang_spec_for_path(path)
        if lang_spec is None:
            return []
        out = []
        for comment in lang_spec.get_prefix_comments(text):
            out.extend(
                StsSelectorLineSpan(
                    start_line_idx=comment.doc_string_node.start_point[0],
                    end_line_idx=comment.doc_string_node.end_point[0] + 1,
                ).get_matching_spans(text, path),
            )
        return out


@dataclass(frozen=True)
class StsSelectorFullFile(StsTextSelector):
    """Selects the entire file"""

    def get_matching_spans(
        self,
        text: str,
        path: str,
    ) -> Sequence[tuple[int, int]]:
        return [(0, len(text))]


@dataclass(frozen=True)
class StsSelectorLineSpan(StsTextSelector):
    """
    Selects a fixed span of lines in the text. Note that it is line
    indicies which are 0 indexed, not 1-indexed line numbers
    """

    start_line_idx: int
    end_line_idx: int
    include_trailing_newline: bool = True

    def get_matching_spans(
        self,
        text: str,
        path: str,
    ) -> Sequence[tuple[int, int]]:
        """Returns the char span corresponding the the line span"""
        lines = text.splitlines(keepends=True)
        if self.start_line_idx > len(lines) or self.end_line_idx > len(lines):
            msg = "Invalid line span selection"
            raise IndexError(msg)
        if self.start_line_idx > self.end_line_idx:
            msg = "Start line index must be less than end line index"
            raise ValueError(msg)
        start_char = sum(len(line) for line in lines[: self.start_line_idx])
        end_char = start_char + sum(
            len(line) for line in lines[self.start_line_idx : self.end_line_idx]
        )
        # Remove trailing '\n' from count if it exists
        if (
            lines[self.end_line_idx - 1].endswith("\n")
            and not self.include_trailing_newline
        ):
            end_char -= 1
        return [(start_char, end_char)]


@dataclass(frozen=True)
class StsVerb:
    """
    A verb applies something at the given selection. This can
    be thought of like 'adding a node of xml'
    """

    @classmethod
    def get_allows_children(cls) -> bool:
        return True

    def operation_is_static(self) -> bool:
        """
        We define the verb operation as "static" if it applies
        once for anything that matches selectors. All the nodes
        in this selection will have the same mark_id.

        The intended usecase here is something like a code question,
        where we are asking about one or more regions of code,
        but really only expect one answer.

        However, if we wanted to say mark all function bodies as
        editable, we would not a static operation. Instead we want
        static to be False, so that we can assign a different value
        to each of the function bodies.
        """
        return False

    @cached_property
    def static_mark_id(self) -> str:
        if not self.operation_is_static():
            msg = "Operation is not static"
            raise ValueError(msg)
        return "s_" + hash_string_base62(
            f"{dataclasses.asdict(self)}",
            digest_size_bytes=4,
        )


@dataclass(frozen=True)
class StsNormalText(StsVerb):
    """
    Normal text is visible to solvers but cannot be by default edited.
    It is the default for any text in the working directory.
    """


@dataclass(frozen=True)
class StsInvisible(StsVerb):
    """Marks text as invisible to a solver, but still present when executing"""

    # TODO alt display text


@dataclass(frozen=True)
class StsPathInvisible(StsVerb):
    """Marks an entire path (and descendants) as invisible to a solver"""


@dataclass(frozen=True)
class StsPlaceTransforming(StsVerb):
    """A verb that specifies an edit to code"""

    max_length_chars: int = None
    max_length_lines: int = None
    lang_spec_name: str = field(
        default_factory=PythonLangSpec().get_lang_md_name,
    )
    """TODO think about if this is actually needed"""
    set_base_space_indent: int = None
    """If set, the solver should assume the value is a at a line-start. The value
    should start with this whitespace."""
    set_ending_text: str = None
    """The solution must end with this string. Solver can make sure
    to either append this or truncate to this point."""
    stop_at_block_end: bool = False
    """Whether the insertion should stop after the end of the current block
    (such as at the end of the current function)"""
    # TODO?
    # stop_text_sequences: list[str] = None
    # """If set, the solver should not insert any text that
    # contains any of these strings. Some solvers might have restrictions
    # on how many stop sequences they support (for example OpenAI only supports
    # up to 4). Note that parameters that `max_line_length` might interact with
    # this as if only limited to one line, then that might use up one of
    # the stop sequences from the solver model."""


@dataclass(frozen=True)
class StsInsert(StsPlaceTransforming):
    """Inserting new text at this point"""

    @classmethod
    def get_allows_children(cls) -> bool:
        return False


@dataclass(frozen=True)
class StsAllowFileWrite(StsPlaceTransforming):
    """
    NOT SUPPORTED YET. Intended to represent allowing the solver to
    write a new file to the working directory
    """

    forced_fn: str = None
    """Whether the solver must use a certain file name. If None, then the solver
    may choose any file name."""
    must_write: bool = False
    """Whether a solution is required to write this file"""
    lang_specs: tuple[LangSpec, ...] = None

    def __post_init__(self):
        raise NotImplementedError


@dataclass(frozen=True)
class StsEditable(StsPlaceTransforming):
    current_text_visible: bool = True


class StsValueQuery(StsVerb):
    """
    An StsVerb that doesn't transform any code in the working directory
    but instead the value itself is part of the answer.
    """

    required: bool

    def operation_is_static(self) -> bool:
        # Default to True for these kinds of queries
        return True


@dataclass(frozen=True)
class StsCodeQuestion(StsValueQuery):
    """
    Requests a freeform answer to question about code. Typically,
    this is not about creating or editing code (those can likely
    be expressed some other way like with StsEditable and instructions),
    but some more natural language-esque question.
    """

    question_text: str
    """The text of the question we want answered"""


@dataclass(frozen=True)
class StsCodeClassificationQuestion(StsCodeQuestion):
    """
    A code question with predefined set of answers.

    A few different values are valid:
    - A json serialized mapping string key answers to a probability.
    - A single string corresponding to selected answer. Assumed to have
        probability 1.0.
    """

    valid_answers: tuple[str, ...]

    def value_to_prob_dict(
        self,
        answer: str,
    ) -> dict[str, float]:
        if "{" in answer and "}" in answer:
            return json.loads(answer)
        return normalize_probs_dict({answer: 1.0}, self.valid_answers)


@dataclass(frozen=True)
class StsDocTestInjectionPoint(StsVerb):
    """
    NOT SUPPORTED YET. A point where a doc test can be inserted
    in cases where we want to support things like toggling on and off
    the test visibility or vary the displayed test cases accross several solves
    """

    def __post_init__(self):
        raise NotImplementedError


@dataclass(frozen=True)
class StsFewShotInjectionPoint(StsVerb):
    """
    TODO some way of representing where to insert few shot examples
    (but need to think through how this fits in with dialog few shots. It
    might be a bit complex than a single point)
    """

    def __post_init__(self):
        raise NotImplementedError


"""
#### Mark Nodes ####
"""


class StsMarkNode(ABC):
    """
    A xml-like (DOM-like) node that represents the result after applying
    a `TransformationSpec` to a file.
    """

    _original_start_char: int
    _original_end_char: int
    _frozen: bool = False

    def __init__(
        self,
        start_char: int,
        end_char: int,
    ):
        self._original_end_char = end_char
        self._original_start_char = start_char

    @property
    def mark_start_original_char(self):
        return self._original_start_char

    @mark_start_original_char.setter
    def mark_start_original_char(self, val: int):
        if self._frozen:
            msg = "Cannot modify frozen mark"
            raise RuntimeError(msg)
        self._original_start_char = val

    @property
    def mark_end_original_char(self):
        return self._original_end_char

    @mark_end_original_char.setter
    def mark_end_original_char(self, val: int):
        if self._frozen:
            msg = "Cannot modify frozen mark"
            raise RuntimeError(msg)
        self._original_end_char = val

    def freeze(self):
        self._frozen = True

    def as_xml(self, have_attributes: bool = True) -> str:
        raise NotImplementedError

    def pretty_print_xml(self) -> None:
        from xml.dom.minidom import parseString

        dom = parseString(self.as_xml())
        pretty_xml_as_string = dom.toprettyxml()
        print(pretty_xml_as_string)

    def depth_first_iterate(self) -> Generator["StsMarkNode", bool, None]:
        """
        A coroutine that yields all nodes starting with self. If
        the coroutine is sent False, then children of the current node are not
        visited.
        """
        visit_children = yield self
        if visit_children is None or visit_children:
            for child in self.children:
                yield from child.depth_first_iterate()

    def iterate_parser_states(
        self,
    ) -> Generator["StandardParserStates", bool, None]:
        """
        A similar coroutine to `depth_first_iterate`, but yields parser states
        that contains information about the current state (such as visible or invisible)
        """
        yield from iterate_parser_states_for_node(self)

    @property
    def children(self):
        raise NotImplementedError

    @children.setter
    def children(self, val: Sequence["StsMarkNode"]):
        raise NotImplementedError


class MarkText(StsMarkNode):
    """
    A DOM text node. Contains the text. For example <a>hello</a> would
    would be an `MarkElement` with a single `MarkText` child with text 'hello'.
    MarkText nodes are always leaves in the DOM tree.
    """

    def __init__(self, text: str, start_char: int, end_char: int):
        super().__init__(start_char, end_char)
        if len(text) != (end_char - start_char):
            msg = (
                f"Text length {len(text)} does not match char span"
                f" {start_char} to {end_char}"
            )
            raise Exception(msg)

        self._text = text

    def split(self, orig_char_point) -> tuple["MarkText", "MarkText"]:
        """Splits this text node into two at the given character index"""
        if (
            orig_char_point < self._original_start_char
            or orig_char_point > self._original_end_char
        ):
            msg = "Split point must be within text node"
            raise ValueError(msg)
        if orig_char_point == self._original_start_char:
            return (None, self)
        if orig_char_point == self._original_end_char:
            return (self, None)
        first = MarkText(
            self._text[: orig_char_point - self._original_start_char],
            self._original_start_char,
            orig_char_point,
        )
        second = MarkText(
            self._text[orig_char_point - self._original_start_char :],
            orig_char_point,
            self._original_end_char,
        )
        return (first, second)

    @property
    def children(self):
        return ()

    @children.setter
    def children(self, val: Sequence["StsMarkNode"]):
        msg = "Cannot set the children of a text node"
        raise RuntimeError(msg)

    def split_span(
        self,
        start,
        end,
    ) -> tuple[Optional["MarkText"], Optional["MarkText"], Optional["MarkText"]]:
        """Returns (before, inner, after) a selection of given span"""
        if end < start:
            msg = "End cannot be less than start."
            raise Exception(msg)

        inner_start_idx = start - self._original_start_char
        inner_end_idx = end - self._original_start_char
        inner_start_idx = max(0, min(inner_start_idx, len(self._text)))
        inner_end_idx = max(0, min(inner_end_idx, len(self._text)))
        if inner_start_idx == 0:
            before = None
        else:
            before = MarkText(
                self._text[:inner_start_idx],
                self._original_start_char,
                self._original_start_char + inner_start_idx,
            )
        if inner_start_idx == inner_end_idx or inner_start_idx >= len(
            self._text,
        ):
            inner = None
        else:
            inner = MarkText(
                self._text[
                    max(0, inner_start_idx) : min(
                        inner_end_idx,
                        len(self._text),
                    )
                ],
                self._original_start_char + max(0, inner_start_idx),
                self._original_start_char + min(inner_end_idx, len(self._text)),
            )
        if inner_end_idx >= len(self._text):
            after = None
        else:
            after = MarkText(
                self._text[min(inner_end_idx, len(self._text)) :],
                self._original_start_char + max(0, inner_end_idx),
                self._original_end_char,
            )
        return before, inner, after

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, val: str):
        if self._frozen:
            msg = "Cannot modify frozen mark"
            raise RuntimeError(msg)
        self._text = val

    def _tuple(self):
        return (self._text, self._original_start_char, self._original_end_char)

    def __eq__(self, other):
        if isinstance(other, MarkText):
            return self._tuple() == other._tuple()
        return NotImplemented

    def as_xml(self, have_attributes: bool = True) -> str:
        return saxutils.escape(self._text)

    def __repr__(self):
        return (
            f"MarkText({self._text!r}, {self._original_start_char},"
            f" {self._original_end_char})"
        )


T = TypeVar("T", bound="StsVerb")


class MarkElement(Generic[T], StsMarkNode):
    """
    A DOM element. It contains a verb (which might have certain properties)
    and a sequence of children.
    """

    def __init__(
        self,
        verb: T,
        children: Sequence[StsMarkNode],
        start_char: int,
        end_char: int,
    ):
        super().__init__(start_char, end_char)
        self._verb = verb
        self._children = children
        self._mark_id = None

    _insert_node_splice_error = (
        "Splitting in the inside a mark element is not supported. "
        "For example <a>foo<b>p</a NOT ALLOWED. Splices the 'b' tag>op</b>. "
        "Maybe in future implementation could be supported by "
        "closing and reopening the tag, but not currently"
    )

    def insert_node(
        self,
        verb: StsVerb,
        start_char: int,
        end_char: int,
    ):
        """
        Inserts a node into this element at a given span. See
        the test cases in `test_transformation_spec.py` for some examples.
        """
        if self._frozen:
            msg = "Cannot modify frozen mark"
            raise RuntimeError(msg)
        if start_char < self.mark_start_original_char:
            msg = "Cannot start before start char"
            raise ValueError(msg)
        if start_char > end_char:
            msg = "Start char cannot be greater than end char"
            raise ValueError(msg)
        if end_char > self._original_end_char:
            msg = "End char cannot be greater than original end char"
            raise ValueError(msg)
        if (
            start_char == self._original_start_char
            and end_char == self._original_end_char
        ):
            self.children = [
                MarkElement(verb, self._children, start_char, end_char),
            ]
            return None
        if len(self._children) == 0:
            msg = "I don't think this should every happen"
            raise RuntimeError(msg)
        new_children_self = []
        # Find the starting child
        for current_original_child_idx, child in enumerate(self._children):
            if start_char >= child.mark_end_original_char:
                new_children_self.append(child)
            else:
                break
        # Now `child` will be set as the child that contains the start of where
        #   we want to split. Now let's split it
        new_node = MarkElement(
            verb,
            [],
            start_char,
            end_char,
        )
        if isinstance(child, MarkText):
            # MarkText won't have more children, so we can just split it directly
            before, inner, after = child.split_span(start_char, end_char)
            if before is not None and start_char < child.mark_end_original_char:
                new_children_self.append(before)
            new_children_self.append(new_node)
            if inner is not None:
                new_node.children.append(inner)
            if after is not None:
                new_children_self.append(after)
            current_original_child_idx += 1  # We now used up the text
        elif isinstance(child, MarkElement):
            new_children_self.append(new_node)
            if end_char <= child.mark_end_original_char:
                if child.mark_start_original_char < start_char:
                    msg = (
                        "Child mark start original character less than start character."
                    )
                    raise Exception(msg)
                child.insert_node(verb, start_char, end_char)
                return None
        else:
            msg = "Unknown mark type"
            raise TypeError(msg)
        # scan to the end for children of our new node
        for child in self._children[current_original_child_idx:]:
            if (
                child.mark_start_original_char < end_char
                and child.mark_end_original_char <= end_char
            ):
                new_node.children.append(child)
            elif (
                child.mark_start_original_char < end_char
                and child.mark_end_original_char > end_char
            ):
                # Need to split
                if isinstance(child, MarkText):
                    before, inner, after = child.split_span(
                        start_char,
                        end_char,
                    )
                    if before is not None:
                        new_node.children.append(before)
                    if inner is not None:
                        new_node.children.append(inner)
                    if after is not None:
                        new_children_self.append(after)
                else:
                    raise ValueError(self._insert_node_splice_error)
            elif child.mark_start_original_char >= end_char:
                new_children_self.append(child)
            else:
                raise ValueError(self._insert_node_splice_error)
        self.children = new_children_self
        return self

    @property
    def verb(self):
        return self._verb

    @verb.setter
    def verb(self, val: StsVerb):
        if self._frozen:
            msg = "Cannot modify frozen mark"
            raise RuntimeError(msg)
        self._verb = val

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, val: Sequence[StsMarkNode]):
        if self._frozen:
            msg = "Cannot modify frozen mark"
            raise RuntimeError(msg)
        self._children = val

    def get_tag(self):
        return str(self.verb.__class__.__name__)

    def get_attributes(self) -> dict[str, str]:
        fields = dataclasses.fields(self.verb)
        out = {}
        for f in fields:
            val = getattr(self.verb, f.name)
            if val is not None:
                out[f.name] = str(val)
        out["_original_start_char"] = str(self._original_start_char)
        out["_original_end_char"] = str(self._original_end_char)
        return out

    def freeze(self):
        super().freeze()
        for child in self._children:
            child.freeze()
        self._children = tuple(self._children)
        return self

    def __eq__(self, other):
        if not isinstance(other, MarkElement):
            return NotImplemented
        return (
            self._verb == other._verb
            and self._original_start_char == other._original_start_char
            and self._original_end_char == other._original_end_char
            and self._children == other._children
        )

    def as_xml(self, have_attributes: bool = True) -> str:
        tag = self.get_tag()
        attributes = {}
        if self._frozen:
            attributes["id"] = self.mark_id
        attributes.update(self.get_attributes())
        if have_attributes:
            attr_str = " ".join(
                f"{key}={saxutils.quoteattr(val)}" for key, val in attributes.items()
            )
        else:
            attr_str = ""
        if len(self._children) == 0:
            return f"<{tag} {attr_str} />"
        children = (child.as_xml() for child in self._children)
        return f"<{tag} {attr_str}>{''.join(children)}</{tag}>"

    def __repr__(self):
        return (
            f"MarkElement({self._verb}, children={self._children}, "
            f"{self._original_start_char}, {self._original_end_char})"
        )

    @property
    def mark_id(self) -> str:
        """Returns a unique string for representing this mark"""
        if self.verb.operation_is_static():
            return self.verb.static_mark_id
        if not self._frozen:
            msg = "Cannot get mark key for unfrozen mark"
            raise RuntimeError(msg)
        # NOTE: This probably could fail with multiple marks on the same character
        if self._mark_id is None:
            self._mark_id = hash_string_base62(
                f"tag:{self.get_tag()},{self._original_start_char},"
                f"{self._original_end_char},{len(self._children)}",
                digest_size_bytes=4,
            )
        return self._mark_id


@dataclass(frozen=True)
class SolveStep:
    """A step for constructing a solution. Writes a value at a certain mark"""

    path: str
    mark_id: MARK_KEY_TYPE
    value: str

    def dict_serialize(self):
        return {
            "path": self.path,
            "mark_id": self.mark_id,
            "value": self.value,
        }


"""
Standard Parser:

Utilities for walking through a markup keeping track of things
like the visibility state.
"""


@dataclass(frozen=True, slots=True)
class StandardParserStates:
    node: "MarkElement" = None
    is_visible: bool = True
    is_editable: bool = False
    parent_state: Optional["StandardParserStates"] = None
    after_visit_children_callbacks: list[callable] = field(default_factory=list)
    """These callables are called after the children of the current node.
    A possible use case is to add a closing tag to the current node.
    It is not clear this is good design but is useful for now."""


def iterate_parser_states_for_node(
    node,
    parent_state: StandardParserStates = None,
) -> Generator[StandardParserStates, bool, None]:
    """
    A similar coroutine to `depth_first_iterate`, but yields parser states
    that contains information about the current state (such as visible or invisible).
    """
    has_parent_state = parent_state is not None
    if not has_parent_state:
        parent_state = StandardParserStates()
    new_visible = parent_state.is_visible
    new_editable = parent_state.is_editable
    if isinstance(node, MarkElement):
        if isinstance(node.verb, StsInvisible):
            new_visible = False
        elif isinstance(node.verb, StsPlaceTransforming):
            new_editable = True
        elif isinstance(node.verb, StsNormalText | StsValueQuery):
            pass
        else:
            msg = f"Unknown verb {node.verb}"
            raise TypeError(msg)
    new_state = StandardParserStates(
        node,
        is_visible=new_visible,
        is_editable=new_editable,
        parent_state=parent_state if has_parent_state else None,
    )
    visit_children = yield new_state
    if visit_children is None or visit_children:
        for child in node.children:
            yield from iterate_parser_states_for_node(child, new_state)
    for listener in new_state.after_visit_children_callbacks:
        listener()



"""
Some basic sandboxing for running untrusted (model generated code).
Not intended to be fully secure.
"""

import logging
import time
logger = logging.getLogger(__name__)
from synthegrator._vendor.epicbox.exceptions import DockerError, EpicBoxError


import multiprocessing
import os
import sys
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Union
import docker
import pytest
import xxhash
from docker.errors import BuildError

from synthegrator._vendor import epicbox
logging.getLogger("synthegrator._vendor.epicbox").setLevel(logging.WARNING)

from synthegrator._vendor.human_eval_ref.execution import (
    create_tempdir,
    reliability_guard,
    swallow_io,
    time_limit,
)
from synthegrator.memory_fs import ProjectDir
from synthegrator.util import pretty_print_python_code

cur_file = Path(__file__).parent.absolute()
IN_GITHUB_ACTION = os.environ.get("GITHUB_ACTIONS") == "true"


@dataclass(frozen=True)
class ExecLimits:
    timeout_cpu_s: int = 60
    timeout_realtime_s: int = 1_000
    memory_limit_mb: int = 4_000
    max_procs: int = 4_000
    networking_allowed: bool = False

    def __getitem__(self, key):
        return getattr(self, key)

    def keys(self):
        return self.__dict__.keys()


@dataclass(frozen=True)
class Cmd:
    cmd: str
    files: tuple[dict[str, bytes | str]] = ()

    @staticmethod
    def from_cmds_str_mix(cmds: Iterable[Union[str, "Cmd"]]) -> list["Cmd"]:
        return [Cmd(cmd) if isinstance(cmd, str) else cmd for cmd in cmds]

    @staticmethod
    def from_cmds_str_mix_tuple(
        *cmds: Iterable[Union[str, "Cmd"]],
    ) -> tuple["Cmd", ...]:
        return tuple(Cmd(cmd) if isinstance(cmd, str) else cmd for cmd in cmds)


@dataclass(frozen=True)
class CmdExecResult:
    cmd: Cmd
    exit_code: int
    completed: bool
    stdout: bytes
    stderr: bytes
    duration: float
    timeout: bool
    oom_killed: bool


@dataclass(frozen=True)
class TestsuiteExecResult:
    __test__ = False
    got_results: bool
    """whether got some xml"""
    collection_error: bool
    """Might be syntax error in the code or runtime error at collection"""
    exec_error: bool
    """The test suite didn't complete (like due to OOM or timeout)"""
    test_suite_exec_result: CmdExecResult
    xml_result: str


_BUILT_IMAGES = {
    # "synthegrator_defects4j",  # TODO: why here
    # "dypy_synthegrator",  # TODO: sort this out. Not clear this should be here
    "NO_DOCKER_IMAGE_BLOCKED",
}

_GITHUB_ACTION_IMAGE_DISALLOWED = {
    "dypy_synthegrator",
}


@dataclass(frozen=True)
class DockerExecutionContext:
    image_name: str
    """The image to try to use or pull"""
    setup_cmds: tuple[Cmd, ...] = ()
    """These are run at the start of execution. Essentially like the CMD in a dockerfile"""
    before_file_copy_cmds: tuple[Cmd, ...] = ()
    """These are run before the working directory is copied into the docker"""
    dockerfile: str = None
    """The string text of a docker file. If provided, should be built when first
    executing code with this environment."""
    build_pre_cmds: tuple[Cmd, ...] = ()
    """These commands run before the dockerfile is built. During build we make a
    temporary directory that serves as the context for the docker. The wd of the
    shell is set to this directory. Thus commands like `cp ~/mydata .` can copy
    into the directory so the dockerfile can access it"""
    default_limits: "ExecLimits" = None

    def __post_init__(self):
        if not (
            all(isinstance(cmd, Cmd) for cmd in self.setup_cmds)
            and all(isinstance(cmd, Cmd) for cmd in self.before_file_copy_cmds)
            and all(isinstance(cmd, Cmd) for cmd in self.build_pre_cmds)
        ):
            msg = "Not all commands are of Cmd type"
            raise TypeError(msg)

    def build(self):
        docker.from_env()
        # _BUILT_IMAGES |= { t for i in client.images.list() for t in i.tags  }
        if self.dockerfile is None:
            if self.image_name not in _BUILT_IMAGES:
                logger.info("Dockerfile is empty. Pulling image: %s", self.image_name)
                raise NotImplementedError("Cannot handle None docker file")
                # TODO: fix pulling images
                # client.images.pull(self.image_name)
        else:
            _build_docker_image(
                self.dockerfile,
                self.image_name,
                self.build_pre_cmds,
            )

    def get_hash(self) -> str:
        if getattr(self, "_hash", None) is None:
            hasher = xxhash.xxh64()
            hasher.update(str(self))
            object.__setattr__(self, "_hash", hasher.hexdigest())
        return self._hash


def make_py_env_with_dependencies(
    name: str = "python_synthegrator",
    base: str = "python:3.11-slim",
    dependencies: list[str] | None = None,
) -> DockerExecutionContext:
    # Auto-detect current pytest version to maintain consistency
    try:
        import pytest
        pytest_version = pytest.__version__
        # Use same minor version to avoid breaking changes
        pytest_spec = f"pytest~={pytest_version}"
    except ImportError:
        # Fallback if pytest not available
        pytest_spec = "pytest~=8.4.1"
    
    docker_file_lines = [
        f"FROM {base}",
        f"RUN pip install --no-cache-dir {pytest_spec}",
    ]
    if dependencies and len(dependencies) > 0:
        dependencies_quoted = [f'"{dep}"' for dep in dependencies]
        dependencies_cmd = (
            f"RUN pip install --no-cache-dir {' '.join(dependencies_quoted)}"
        )
        docker_file_lines.append(dependencies_cmd)
    docker_file_lines.append("WORKDIR /app")
    return DockerExecutionContext(
        image_name=name,
        dockerfile="\n".join(docker_file_lines),
    )


# fmt: off
PY_DEFAULT_DOCKER_ENV = make_py_env_with_dependencies(
    name="python_synthegrator",
    dependencies=["numpy==2.1.0"]
)
# fmt: on


JAVA_DEFAULT_DOCKER_ENV = DockerExecutionContext(
    image_name="java_synthegrator",
    build_pre_cmds=Cmd.from_cmds_str_mix_tuple(
        f"cp {cur_file / 'lang_specs/extra_files/default_java_pom.xml'} pom.xml",
        f"cp {cur_file / 'lang_specs/extra_files/DummyTest.java'} DummyTest.java",
    ),
    dockerfile=(
        # fmt: off
        "FROM maven:3.9.4-eclipse-temurin-17\n"
        "WORKDIR /app\n"
        "COPY pom.xml .\n"
        # Copy in a dummy file that uses Junit and force it install maven dependencies.
        #   Makeing the dependencies part of the container is important so that we
        #   aren't redownloading the dependencies every time we evaluate a problem.
        "COPY DummyTest.java ./src/test/java/DummyTest.java\n"
        "RUN mvn dependency:go-offline\n"
        "RUN mvn dependency:resolve-plugins\n"
        "RUN mvn test\n"
        # fmt: on
    ),
    setup_cmds=Cmd.from_cmds_str_mix_tuple(
        "cp /app/pom.xml .",
    ),
)

_dypy_root = Path(__file__).parent / "synthdatasets/dypybench_data/dypy_run_scripts"
DYPYBENCH_DOCKER_ENV = DockerExecutionContext(
    image_name="dypy_synthegrator",
    dockerfile=(
        # fmt: off
        "FROM dypybench/dypybench:v1.0\n"
        "COPY dypybench.py.nolint /DyPyBench/dypybench.py\n"
        "COPY run-test-temp.sh /DyPyBench/scripts/run-test-temp.sh\n"
        "COPY copy-project.sh /DyPyBench/scripts/copy-project.sh\n"
        "RUN chmod +x /DyPyBench/scripts/run-test-temp.sh /DyPyBench/scripts/copy-project.sh\n"
        # fmt: on
    ),
    build_pre_cmds=Cmd.from_cmds_str_mix_tuple(
        f"cp {_dypy_root / 'run-test-temp.sh'} .",
        f"cp {_dypy_root / 'copy-project.sh'} .",
        f"cp {_dypy_root / 'dypybench.py.nolint'} .",
    ),
    default_limits=ExecLimits(
        timeout_cpu_s=60 * 60,
        timeout_realtime_s=60 * 60,
        networking_allowed=True,
    ),
)


def run_on_docker(
    docker_context: DockerExecutionContext,
    cmds: Iterable[Cmd | str],  # Commands to run after starting
    files: ProjectDir = None,  # files to copy over
    limits: ExecLimits | None = None,
    interactive_shell_debug: bool = False,
    retries_on_docker_error: int = 2,
) -> list["CmdExecResult"]:
    """Run a set of commands in a docker container. Returns a list of results."""
    # TODO: This is just a hack to get this working. Will
    #  need to benchmark and optimize this later. In particular it restarts
    #  the container multiple time for each command, and it should just keep
    #  it running.
    # TODO: migrate away from epicbox. Epicbox has weird quirk where not only
    #  is it restarting every time, but it ignores the working directory in the
    #  the dockerfile and is instead using a /sandbox volume
    cmds = Cmd.from_cmds_str_mix(cmds)
    if limits is None:
        limits = docker_context.default_limits or ExecLimits()

    for attempt in range(retries_on_docker_error):
        try:
            docker_context.build()
            break
        except DockerError as e:
            logger.warning(f"Docker error: {e}.\n Retrying {attempt + 1} of {retries_on_docker_error}...")
            time.sleep(2 ** attempt)

    profile_name = f"synthegrator_{docker_context.image_name}"
    epicbox.configure(
        profiles=[
            epicbox.Profile(
                profile_name,
                docker_context.image_name,
                network_disabled=not limits.networking_allowed,
            ),
        ],
    )

    limits_d = {
        "cputime": limits.timeout_cpu_s,
        "realtime": limits.timeout_realtime_s,
        "memory": limits.memory_limit_mb,
        "processes": limits.max_procs,
        # TODO: add a way to limit disk space
    }

    for attempt in range(retries_on_docker_error):
        try:
            with epicbox.working_directory() as wd:
                results = []

                def run_cmd(cmd: Cmd):
                    if len(cmd.files) > 0 and not all(
                        isinstance(f["name"], str) and isinstance(f["content"], bytes)
                        for f in cmd.files
                    ):
                        msg = "Files in cmd are not str/bytes pairs."
                        raise Exception(msg)

                    run = epicbox.run(
                        profile_name,
                        cmd.cmd,
                        workdir=wd,
                        limits=limits_d,
                        files=cmd.files,
                    )
                    results.append(
                        CmdExecResult(
                            cmd=cmd,
                            completed=not (run["timeout"] or run["oom_killed"]),
                            **run,
                        ),
                    )

                # Pre copy
                if docker_context.before_file_copy_cmds:
                    for cmd in docker_context.before_file_copy_cmds:
                        run_cmd(cmd)

                # Copy over working directory files
                if files:
                    # hackily move the files over. Have to work around how epicbox
                    #  handles files where can only have flat set of files.
                    transfer_tar_name = "transfer_data_to_val.tar"
                    epic_files = [
                        {
                            "name": transfer_tar_name,
                            "content": files.convert_to_tar(gzip=False),
                        },
                    ]
                    epicbox.run(
                        profile_name,
                        f"tar -xf {transfer_tar_name}; rm {transfer_tar_name}",
                        workdir=wd,
                        files=epic_files,
                    )

                # Run rest of commands
                for cmd in [*docker_context.setup_cmds, *cmds]:
                    run_cmd(cmd)

                if interactive_shell_debug:
                    while cmd := input("cmd: "):
                        r = epicbox.run(profile_name, cmd, workdir=wd, limits=limits_d)
                        print(r)
                        print(r["stdout"].decode("utf-8"))
                        print(r["stderr"].decode("utf-8"), file=sys.stderr)
            return results
        except (DockerError, EpicBoxError) as e:
            logger.warning(f"Docker error: {e}.\n Retrying {attempt + 1} of {retries_on_docker_error}...")
            time.sleep(2 ** attempt)
            continue
    raise Exception("Failed to run commands on docker")


def convert_pytest_outputs_to_testsuite_exec_result(
    pytest_out: CmdExecResult,
    cat_out: CmdExecResult,
) -> TestsuiteExecResult:
    if not cat_out.cmd.cmd.startswith("cat "):
        msg = "Output command does not start with cat"
        raise Exception(msg)

    pytest_exit_code = pytest_out.exit_code
    ran_tests = pytest_out.completed and pytest_exit_code in {
        pytest.ExitCode.OK.value,
        pytest.ExitCode.NO_TESTS_COLLECTED.value,
        pytest.ExitCode.TESTS_FAILED.value,
        # pytest.ExitCode.INTERRUPTED.value,
        # ^ This for some reason will happen if have a syntax error in the test
        # internal error, or pytest command usage error
    }
    xml_result = cat_out.stdout.decode("utf-8")
    ran_with_results = len(xml_result) > 0 and ran_tests
    return TestsuiteExecResult(
        got_results=ran_with_results,
        collection_error=pytest_out.completed
        and (
            pytest_exit_code
            in {
                pytest.ExitCode.USAGE_ERROR.value,
                pytest.ExitCode.INTERNAL_ERROR.value,
                pytest.ExitCode.INTERRUPTED.value,
            }
        ),
        exec_error=not pytest_out.completed,
        test_suite_exec_result=pytest_out,
        xml_result=xml_result,
    )


def pytest_on_docker(
    files: ProjectDir,
    run_file: str,
    docker_env: DockerExecutionContext = PY_DEFAULT_DOCKER_ENV,
    extra_setup_cmds: Iterable[Cmd] = (),
    limits: ExecLimits = None,  # Use whatever docker_env has, or default
) -> TestsuiteExecResult:
    xml_name = "test_report_results_from_sb.xml"
    if limits is None:
        limits = docker_env.default_limits or ExecLimits()
    out = run_on_docker(
        docker_env,
        (
            *docker_env.setup_cmds,
            *extra_setup_cmds,
            *[
                Cmd(f"rm {xml_name}"),
                Cmd(f"python -m pytest {run_file} -s --junitxml={xml_name}"),
                Cmd(f"cat {xml_name}"),
            ],
        ),
        files=files,
        interactive_shell_debug=False,
        limits=limits,
    )
    return convert_pytest_outputs_to_testsuite_exec_result(out[-2], out[-1])


def _build_docker_image(dockerfile_str, tag, pre_cmds: Iterable[Cmd]):
    if IN_GITHUB_ACTION and tag in _GITHUB_ACTION_IMAGE_DISALLOWED:
        raise RuntimeError(f"Building docker image {tag} is disabled in GitHub Actions")
    client = docker.from_env()

    bult_images = {t for i in client.images.list() for t in i.tags}

    if tag in _BUILT_IMAGES | bult_images:
        return None
    logger.info(f"Building docker image {tag}...")
    # Create a temporary directory
    with TemporaryDirectory() as tmp_dir_obj:
        tmpdir = Path(tmp_dir_obj)
        # Write Dockerfile string to a file in the temporary directory
        (tmpdir / "Dockerfile").write_text(dockerfile_str)

        for cmd in pre_cmds:
            cd_cmd = f"cd {tmpdir} && {cmd.cmd}"
            os.system(cd_cmd)

        # Build the docker image from the Dockerfile in the temporary directory
        try:
            image, build_logs = client.images.build(
                path=str(tmpdir),
                tag=tag,
                rm=True,
            )
        except BuildError as e:
            for chunk in e.build_log:
                if "stream" in chunk:
                    for line in chunk["stream"].splitlines():
                        logger.error(line)
            raise
        _BUILT_IMAGES.add(tag)
        logger.info(f"docker image {tag} build complete")
        return (image, build_logs)


##############################################
# NON-DOCKER VERSION ###


_result_var_name = "my_exec_result_value"


def combine_code_and_result_expression(
    code: str,
    result_expression: str,
) -> str:
    return (
        code
        + "\n\n"
        + "#### Result Expression ####\n"
        + _result_var_name
        + " = "
        + result_expression
    )


def safer_exec(
    code,
    result_expression,
    timeout=10.0,
    user_confirmation=True,
) -> "ExecResult":
    """
    DEPRECATED: switching to docker running instead.
    Runs code in the current python process with just some more risky globals (like file
    writing) set to None. This is based off how HumanEval ran their code. It is not
    intended to be fully secure.
    """
    if user_confirmation:
        logger.info("About to exec")
        logger.info("---")
        pretty_print_python_code(
            combine_code_and_result_expression(code, result_expression),
        )
        logger.info("---")
    # Ask the user to confirm
    if (
        user_confirmation
        and input("Are you sure you want to exec this code? (y/n)") != "y"
    ):
        msg = "User did not confirm"
        raise Exception(msg)

    result_queue = multiprocessing.Queue()

    # Create a Process with your_function as target and pass the argument and queue
    process = multiprocessing.Process(
        target=_unsafe_execute,
        args=(result_queue, code, result_expression, timeout),
    )

    # Start the process
    process.start()

    # Get the result from the queue
    result = result_queue.get(timeout=timeout + 1)

    # Join the process
    process.join(timeout=1)
    if process.is_alive():
        process.kill()

    # Only show executed code if it failed due to a type or assertion error
    # TODO make configurable?
    if isinstance(result.exception, AssertionError | TypeError):
        pretty_print_python_code(
            combine_code_and_result_expression(code, result_expression),
        )
        logger.info("Result: %s", result)

    return result


def _unsafe_execute(q, code, result_expression: str, timeout):
    # Copied from HumanEval. DEPRECATED for running on docker
    with create_tempdir():
        # These system calls are needed when cleaning up tempdir.
        import os
        import shutil

        rmtree = shutil.rmtree
        rmdir = os.rmdir
        chdir = os.chdir

        # Disable functionalities that can make destructive changes to the test.
        reliability_guard()
        check_program = combine_code_and_result_expression(
            code,
            result_expression,
        )

        try:
            exec_globals = {"my_exec_result_value": None}
            # track the start of the execution
            import time

            start_time = time.time()

            with swallow_io(), time_limit(timeout):
                # WARNING
                # This program exists to execute untrusted model-generated code. Although
                # it is highly unlikely that model-generated code will do something overtly
                # malicious in response to this test suite, model-generated code may act
                # destructively due to a lack of model capability or alignment.
                # Users are strongly encouraged to sandbox this evaluation suite so that it
                # does not perform destructive actions on their host or network. For more
                # information on how OpenAI sandboxes its code, see the accompanying paper.
                # Once you have read this disclaimer and taken appropriate precautions,
                # uncomment the following line and proceed at your own risk
                exec(check_program, exec_globals)
            q.put(
                ExecResult(
                    output=exec_globals[_result_var_name],
                    completed=True,
                    runtime=time.time() - start_time,
                    exception=None,
                ),
            )
        except TimeoutError as e:
            q.put(
                ExecResult(
                    output=None,
                    completed=False,
                    runtime=time.time() - start_time,
                    exception=e,
                ),
            )
        except BaseException as e:
            q.put(
                ExecResult(
                    output=None,
                    completed=False,
                    runtime=time.time() - start_time,
                    exception=e,
                ),
            )

        # Needed for cleaning up.
        shutil.rmtree = rmtree
        os.rmdir = rmdir
        os.chdir = chdir


@dataclass
class ExecResult:
    # Deprecated with docker running
    output: Any
    completed: bool
    runtime: float
    exception: BaseException | None = None



import itertools
from unittest.mock import MagicMock

from synthegrator.code_solver import DummyCodeSolverAutoRegressive
from synthegrator.execution_threading import evaluate_all_solutions, solve_and_evaluate_problems
from synthegrator.solution_eval import SolutionEvaluator
from synthegrator.synthdatasets.human_eval import yield_human_eval


class FailEvaluator(SolutionEvaluator):
    def evaluate(self, solution):
        raise ValueError("This evaluator always fails")


def test_catch_fail_with_trace():
    solver = DummyCodeSolverAutoRegressive(MagicMock())
    problems = itertools.islice(yield_human_eval(), 10)
    evals = solve_and_evaluate_problems(
        solver,
        problems,
        max_threads_solve=1,
        max_threads_eval=2,
        evaluator=FailEvaluator(None),
    )
    evals = list(evals)
    assert len(evals) == 10
    assert evals[0].main_metric is None
    assert evals[0].exception is not None
    assert str(evals[0].exception) == "This evaluator always fails"
    assert evals[0].exception.__traceback__ is not None
    assert evals[0].exception_traceback is not None
    assert __file__ in evals[0].exception_traceback